import numpy as np

from ..field import make_hexagonal_grid, Field
from .generic import make_elliptical_aperture, make_spider, make_circular_aperture, make_hexagonal_aperture, make_segmented_aperture, make_shifted_aperture, make_spider_infinite, make_obstructed_circular_aperture, make_rectangular_aperture, make_obstruction, make_regular_polygon_aperture, make_irregular_polygon_aperture

import functools

_vlt_telescope_aliases = {'antu': 'ut1', 'kueyen': 'ut2', 'melipal': 'ut3', 'yepun': 'ut4'}

def make_vlt_aperture(
        normalized=False, telescope='ut3', with_spiders=True, with_M3_cover=False,
        return_segments=False):
    '''Make the VLT aperture.

    This aperture is based on the ERIS pupil documentation: VLT-SPE-AES-11310-0006.

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 8.0 meters for UT1-3 and 8.1196 meters for UT4.
    telescope : one of {'ut1', 'ut2', 'ut3', 'ut4', 'antu', 'kueyen', 'melipal', 'yepun'}
        The specific telescope on the VLT, case insensitive. Default: UT3.
    with_spiders : boolean
        If this is False, the spiders will be left out. Default: True.
    with_M3_cover : boolean
        If this is True, a cover will be created for the M3 in stowed position.
        This M3 cover is only available on UT4, mimicking the ERIS pupil. A warning
        will be emitted when using an M3 cover with other UTs. Default: False.
    return_segments : boolean
        If this is True, the pupil quadrants (segments) will also be returned.

    Returns
    -------
    aperture : Field generator
        The VLT aperture.
    segments : list of Field generators
        The segments. Only returned when `return_segments` is True.
    '''
    telescope = telescope.lower()
    if telescope not in ['ut1', 'ut2', 'ut3', 'ut4']:
        if telescope not in _vlt_telescope_aliases:
            raise ValueError(f'The VLT telescope "{telescope}" has not been implemented.')
        telescope = _vlt_telescope_aliases[telescope]

    if telescope in ['ut1', 'ut2', 'ut3']:
        pupil_diameter = 8.0  # meter
        central_obscuration_ratio = 1.116 / pupil_diameter
    elif telescope == 'ut4':
        pupil_diameter = 8.1196  # meter
        central_obscuration_ratio = 0.6465 * 2 / pupil_diameter

    spider_width = 0.040  # meter
    spider_offset = 0.4045  # meter
    spider_outer_radius = 4.2197  # meter
    outer_diameter_M3_stow = 1.070  # meter
    angle_between_spiders = 101  # degrees

    if with_M3_cover and telescope != 'ut4':
        import warnings
        warnings.warn('Using the M3 cover on a telescope other than UT4 is not realistic.', stacklevel=2)

    if normalized:
        spider_width /= pupil_diameter
        spider_offset /= pupil_diameter
        spider_outer_radius /= pupil_diameter
        outer_diameter_M3_stow /= pupil_diameter
        pupil_diameter = 1.0

    obstructed_aperture = make_obstructed_circular_aperture(pupil_diameter, central_obscuration_ratio)

    if with_spiders or return_segments:
        spider_inner_radius = spider_offset / np.cos(np.radians(45 - (angle_between_spiders - 90) / 2))

        spider_start_1 = -spider_inner_radius * np.array([np.cos(np.pi / 4), np.sin(np.pi / 4)])
        spider_end_1 = spider_outer_radius * np.array([np.cos(np.pi), np.sin(np.pi)])

        spider_start_2 = -spider_inner_radius * np.array([np.cos(np.pi / 4), np.sin(np.pi / 4)])
        spider_end_2 = spider_outer_radius * np.array([np.cos(-np.pi / 2), np.sin(-np.pi / 2)])

        spider_start_3 = spider_inner_radius * np.array([np.cos(np.pi / 4), np.sin(np.pi / 4)])
        spider_end_3 = spider_outer_radius * np.array([np.cos(0), np.sin(0)])

        spider_start_4 = spider_inner_radius * np.array([np.cos(np.pi / 4), np.sin(np.pi / 4)])
        spider_end_4 = spider_outer_radius * np.array([np.cos(np.pi / 2), np.sin(np.pi / 2)])

    if with_spiders:
        spider1 = make_spider(spider_start_1, spider_end_1, spider_width)
        spider2 = make_spider(spider_start_2, spider_end_2, spider_width)
        spider3 = make_spider(spider_start_3, spider_end_3, spider_width)
        spider4 = make_spider(spider_start_4, spider_end_4, spider_width)

    if with_M3_cover:
        m3_cover = make_obstruction(make_rectangular_aperture(outer_diameter_M3_stow, center=[outer_diameter_M3_stow / 2, 0]))

    if with_spiders:
        if with_M3_cover:
            def func(grid):
                return Field(obstructed_aperture(grid) * spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid) * m3_cover(grid), grid)
        else:
            def func(grid):
                return Field(obstructed_aperture(grid) * spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid), grid)
    else:
        if with_M3_cover:
            def func(grid):
                return Field(obstructed_aperture(grid) * m3_cover(grid), grid)
        else:
            def func(grid):
                return Field(obstructed_aperture(grid), grid)

    if return_segments:
        n1 = (spider_end_1[1] - spider_start_1[1], spider_start_1[0] - spider_end_1[0])
        c1 = np.dot(n1, spider_end_1)

        n2 = (spider_end_2[1] - spider_start_2[1], spider_start_2[0] - spider_end_2[0])
        c2 = np.dot(n2, spider_end_2)

        n3 = (spider_end_3[1] - spider_start_3[1], spider_start_3[0] - spider_end_3[0])
        c3 = np.dot(n3, spider_end_3)

        n4 = (spider_end_4[1] - spider_start_4[1], spider_start_4[0] - spider_end_4[0])
        c4 = np.dot(n4, spider_end_4)

        ns = [n1, n4, n3, n2]
        cs = [c1, c4, c3, c2]

        def segment(n1, c1, n2, c2, grid):
            if grid.is_separated:
                x, y = grid.separated_coords
                f = (np.dot(n1, np.array([x[np.newaxis, :], y[:, np.newaxis]], dtype = object)) > c1) * 1.0
                f *= (np.dot(n2, np.array([x[np.newaxis, :], y[:, np.newaxis]], dtype = object)) < c2) * 1.0
                intersection = np.array([c1, c2]).dot(np.linalg.inv(np.array([n1, n2])))
                ni = np.array([-intersection[1], -intersection[0]])
                f *= (np.dot(ni, np.array([x[np.newaxis, :], y[:, np.newaxis]], dtype = object)) < 0) * 1.0
            else:
                x, y = grid.coords
                f = (np.dot(n1, np.array([x, y])) > c1) * 1.0
                f *= (np.dot(n2, np.array([x, y])) < c2) * 1.0
                intersection = np.array([c1, c2]).dot(np.linalg.inv(np.array([n1, n2])))
                ni = np.array([-intersection[1], -intersection[0]])
                f *= (np.dot(ni, np.array([x, y])) < 0) * 1.0
            f = f.ravel()
            f *= func(grid)

            if with_M3_cover:
                f *= m3_cover(grid)
            return Field(f.astype('float'), grid)

        segments = []
        for i in range(4):
            segments.append(functools.partial(
                segment, np.roll(ns, -i, axis=0)[0], np.roll(cs, -i, axis=0)[0],
                np.roll(ns, -i, axis=0)[1], np.roll(cs, -i, axis=0)[1]))

    if return_segments:
        return func, segments
    else:
        return func

def make_vlti_aperture(zenith_angle=0, azimuth=0, with_spiders=True, return_segments=False):
    '''Make the VLTI aperture for interferometry.

    The position of each VLT is taken from the VLTI user manual: VLT-MAN-ESO-15000-4552.
    This function does not apply any differential piston.

    Parameters
    ----------
    zenith_angle : scalar
        The zenith angle component of the pointing direction.
    azimuth : scalar
        The azimuth component of the pointing direction. Defined from north to east.
    with_spiders : boolean
        Include the secondary mirror support structure in the aperture.
    return_segments : boolean
        If this is True, the segments will also be returned as a ModeBasis.

    Returns
    -------
    Field generator
        The VLTI aperture.
    segments : list of Field generators
        The individual telescopes. Only returned when `return_segments` is True.
    '''
    # UT1 is taken as a reference
    baseline_UT12 = np.array([24.8, 50.8])
    baseline_UT13 = np.array([54.8, 86.5])
    baseline_UT14 = np.array([113.2, 64.3])

    relative_position = np.array([[0, 0], baseline_UT12, baseline_UT13, baseline_UT14])

    # Calculate the middle between the extremes of the 4 telescopes
    reference_position = (np.max(relative_position, axis=0) + np.min(relative_position, axis=0)) / 2
    telescope_positions = np.array([-reference_position, baseline_UT12 - reference_position, baseline_UT13 - reference_position, baseline_UT14 - reference_position])

    # Make the new local coordinate basis from the pointing vector
    def make_point_vector(theta, phi):
        return np.array([np.sin(phi) * np.sin(theta), np.cos(phi) * np.sin(theta), np.cos(theta)])

    w = make_point_vector(zenith_angle, azimuth)
    v = make_point_vector(zenith_angle + np.pi / 2, azimuth)
    u = np.cross(v, w)

    telescope_apertures = []
    for telescope_name, position in zip(['ut1', 'ut2', 'ut3', 'ut4'], telescope_positions):
        u_position = u[0] * position[0] + u[1] * position[1]
        v_position = v[0] * position[0] + v[1] * position[1]

        single_ut = make_shifted_aperture(make_vlt_aperture(telescope=telescope_name, with_spiders=with_spiders), shift=[u_position, v_position])
        telescope_apertures.append(single_ut)

    def func(grid):
        res = 0
        for ap in telescope_apertures:
            res += ap(grid)
        return res

    if return_segments:
        return func, telescope_apertures
    else:
        return func

def make_vlti_dopd_map(zenith_angle=0, azimuth=0, with_spiders=True, return_segments=False):
    '''Make the VLTI differential OPD map for interferometry.

    The position of each VLT is taken from the VLTI user manual: VLT-MAN-ESO-15000-4552.

    Parameters
    ----------
    zenith_angle : scalar
        The zenith angle component of the pointing direction.
    azimuth : scalar
        The azimuth component of the pointing direction. Defined from north to east.
    with_spiders : boolean
        Include the secondary mirror support structure in the aperture.
    return_segments : boolean
        If this is True, the segments will also be returned as a ModeBasis.

    Returns
    -------
    Field generator
        The full VLTI dOPD map.
    segments : list of Field generators
        The individual dOPD maps for each telescope. Only returned when `return_segments` is True.
    '''
    # UT1 is taken as a reference
    baseline_UT12 = np.array([24.8, 50.8])
    baseline_UT13 = np.array([54.8, 86.5])
    baseline_UT14 = np.array([113.2, 64.3])

    relative_position = np.array([[0, 0], baseline_UT12, baseline_UT13, baseline_UT14])

    # Calculate the middle between the extremes of the 4 telescopes
    reference_position = (np.max(relative_position, axis=0) + np.min(relative_position, axis=0)) / 2
    telescope_positions = np.array([-reference_position, baseline_UT12 - reference_position, baseline_UT13 - reference_position, baseline_UT14 - reference_position])

    # Make the new local coordinate basis from the pointing vector
    def make_point_vector(theta, phi):
        return np.array([np.sin(phi) * np.sin(theta), np.cos(phi) * np.sin(theta), np.cos(theta)])

    w = make_point_vector(zenith_angle, azimuth)
    v = make_point_vector(zenith_angle + np.pi / 2, azimuth)
    u = np.cross(v, w)

    dOPD = []
    telescope_apertures = []
    for telescope_name, position in zip(['ut1', 'ut2', 'ut3', 'ut4'], telescope_positions):
        u_position = u[0] * position[0] + u[1] * position[1]
        v_position = v[0] * position[0] + v[1] * position[1]
        dOPD_i = w[0] * position[0] + w[1] * position[1]
        dOPD.append(dOPD_i)

        single_ut = make_shifted_aperture(make_vlt_aperture(telescope=telescope_name, with_spiders=with_spiders), shift=[u_position, v_position])

        telescope_apertures.append(single_ut)

    def ut_with_opd(ut, dOPD_i):
        return lambda grid: ut(grid) * dOPD_i
    telescope_apertures = [ut_with_opd(ut, dOPD_i) for ut, dOPD_i in zip(telescope_apertures, dOPD)]

    def func(grid):
        res = 0
        for ap in telescope_apertures:
            res += ap(grid)
        return res

    if return_segments:
        return func, telescope_apertures
    else:
        return func

def make_subaru_aperture():
    pass

def make_lbt_aperture():
    pass

def make_magellan_aperture(normalized=False, with_spiders=True):
    '''Make the Magellan aperture.

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 6.5 meters.
    with_spiders: boolean
        If this is False, the spiders will be left out.

    Returns
    -------
    Field generator
        The Magellan aperture.
    '''
    pupil_diameter = 6.5  # meter
    spider_width1 = 0.75 * 0.0254  # meter
    spider_width2 = 1.5 * 0.0254  # meter
    central_obscuration_ratio = 0.29
    spider_offset = [0, 0.34]  # meter

    if normalized:
        spider_width1 /= pupil_diameter
        spider_width2 /= pupil_diameter
        spider_offset = [x / pupil_diameter for x in spider_offset]
        pupil_diameter = 1.0

    spider_offset = np.array(spider_offset)

    mirror_edge1 = (pupil_diameter / (2 * np.sqrt(2)), pupil_diameter / (2 * np.sqrt(2)))
    mirror_edge2 = (-pupil_diameter / (2 * np.sqrt(2)), pupil_diameter / (2 * np.sqrt(2)))
    mirror_edge3 = (pupil_diameter / (2 * np.sqrt(2)), -pupil_diameter / (2 * np.sqrt(2)))
    mirror_edge4 = (-pupil_diameter / (2 * np.sqrt(2)), -pupil_diameter / (2 * np.sqrt(2)))

    obstructed_aperture = make_obstructed_circular_aperture(pupil_diameter, central_obscuration_ratio)

    if not with_spiders:
        return obstructed_aperture

    spider1 = make_spider(spider_offset, mirror_edge1, spider_width1)
    spider2 = make_spider(spider_offset, mirror_edge2, spider_width1)
    spider3 = make_spider(-spider_offset, mirror_edge3, spider_width2)
    spider4 = make_spider(-spider_offset, mirror_edge4, spider_width2)

    def func(grid):
        return obstructed_aperture(grid) * spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid)
    return func

def make_hale_aperture(normalized=False, with_spiders=True):
    '''Make the Hale Telescope aperture.

    The size of the spiders and the attachements were based on the actual pupil measurements from Figure 2 of [Galicher2019]_

    .. [Galicher2019] Raphael Galicher et al., "Minimization of non-common path aberrations at the Palomar telescope using a self-coherent camera." Astronomy & Astrophysics 631 (2019): A143.

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 5.08 meters.
    with_spiders: boolean
        If this is False, the spiders will be left out.

    Returns
    -------
    Field generator
        The Hale Telescope aperture.
    '''
    pupil_diameter = 5.08  # meter
    central_obscuration_diameter = 1.86  # meter
    spider_width = 2 * 0.024  # meter
    central_obscuration_ratio = central_obscuration_diameter / pupil_diameter

    # Attachement points of the spiders at the central obscuration
    box_heigth = 2 * 0.06
    box_width = 2 * 0.0932 + central_obscuration_diameter

    if normalized:
        spider_width /= pupil_diameter
        box_heigth /= pupil_diameter
        box_width /= pupil_diameter
        pupil_diameter = 1.0

    box1 = make_rectangular_aperture([box_width, box_heigth])
    box2 = make_rectangular_aperture([box_heigth, box_width])

    if not with_spiders:
        obstructed_aperture = make_obstructed_circular_aperture(pupil_diameter, central_obscuration_ratio)
    else:
        obstructed_aperture = make_obstructed_circular_aperture(pupil_diameter, central_obscuration_ratio, num_spiders=4, spider_width=spider_width)

    def func(grid):
        return Field(obstructed_aperture(grid) * (1 - box1(grid)) * (1 - box2(grid)), grid)

    return func

def make_luvoir_a_aperture(
        normalized=False, with_spiders=True, with_segment_gaps=True,
        gap_padding=1, segment_transmissions=1, return_header=False, return_segments=False):
    '''Make the LUVOIR A aperture.

    This aperture changes frequently. This one is based on LUVOIR Apertures dimensions
    from Matt Bolcar, LUVOIR lead engineer (as of 10 April 2019)
    Spiders and segment gaps can be included or excluded, and the transmission for each
    of the segments can also be changed. Segments can be returned as well.

    Parameters
    ----------
    normalized : boolean
        If this is True, the pupil diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 15.0 meters.
    with_spiders : boolean
        Include the secondary mirror support structure in the aperture.
    with_segment_gaps : boolean
        Include the gaps between individual segments in the aperture.
    gap_padding : scalar
        Arbitrary padding of gap size to represent gaps on smaller arrays - this effectively
        makes the gaps larger and the segments smaller to preserve the same segment pitch.
    segment_transmissions : scalar or array_like
        The transmission for each of the segments. If this is a scalar, this transmission
        will be used for all segments.
    return_header : boolean
        If this is True, a header will be returned giving all important values for the
        created aperture for reference.
    return_segments : boolean
        If this is True, the segments will also be returned as a ModeBasis.

    Returns
    -------
    aperture : Field generator
        The LUVOIR A aperture.
    aperture_header : dict
        A dictionary containing all quantities used when making this aperture. Only returned if
        `return_header` is True.
    segments : list of Field generators
        The segments. Only returned when `return_segments` is True.
    '''
    pupil_diameter = 15.0  # m actual circumscribed diameter, used for lam/D calculations other measurements normalized by this diameter
    pupil_inscribed = 13.5  # m actual inscribed diameter
    actual_segment_flat_diameter = 1.2225  # m actual segment flat-to-flat diameter
    actual_segment_gap = 0.006  # m actual gap size between segments
    spider_width = 0.150  # m actual strut size
    spid_start = 0.30657  # m spider starting point distance from center of aperture
    num_rings = 6  # number of full rings of hexagons around central segment
    lower_spider_angle = 12.7  # deg spiders are upside-down 'Y' shaped; degree the lower two spiders are offset from vertical by this amount

    # padding out the segmentation gaps so they are visible and not sub-pixel
    segment_gap = actual_segment_gap * gap_padding
    if not with_segment_gaps:
        segment_gap = 0

    segment_flat_diameter = actual_segment_flat_diameter - (segment_gap - actual_segment_gap)
    segment_circum_diameter = 2 / np.sqrt(3) * segment_flat_diameter

    if not with_segment_gaps:
        segment_gap = 0

    aperture_header = {
        'TELESCOP': 'LUVOIR A',
        'D_CIRC': pupil_diameter,
        'D_INSC': pupil_inscribed,
        'SEG_F2F_D': actual_segment_flat_diameter,
        'SEG_GAP': actual_segment_gap,
        'STRUT_W': spider_width,
        'STRUT_AN': lower_spider_angle,
        'NORM': normalized,
        'SEG_TRAN': segment_transmissions,
        'GAP_PAD': gap_padding,
        'STRUT_ST': spid_start,
        'PROV': 'MBolcar ppt 20180815'
    }

    if normalized:
        segment_circum_diameter /= pupil_diameter
        actual_segment_flat_diameter /= pupil_diameter
        actual_segment_gap /= pupil_diameter
        spider_width /= pupil_diameter
        spid_start /= pupil_diameter
        pupil_diameter = 1.0

    segment_positions = make_hexagonal_grid(actual_segment_flat_diameter + actual_segment_gap, num_rings)

    # clipping the "corner" segments of the outermost rings
    segment_positions = segment_positions.subset(make_circular_aperture(pupil_diameter * 0.98))
    segment_positions = segment_positions.subset(lambda grid: ~(make_circular_aperture(segment_circum_diameter)(grid) > 0))

    segment = make_hexagonal_aperture(segment_circum_diameter, np.pi / 2)

    if with_spiders:
        spider1 = make_spider_infinite([0, 0], 90, spider_width)
        spider2 = make_spider_infinite([spid_start, 0], 270 - lower_spider_angle, spider_width)
        spider3 = make_spider_infinite([-spid_start, 0], 270 + lower_spider_angle, spider_width)

    segmented_aperture = make_segmented_aperture(segment, segment_positions, segment_transmissions, return_segments=return_segments)

    if return_segments:
        segmented_aperture, segments = segmented_aperture

    def func(grid):
        res = segmented_aperture(grid)

        if with_spiders:
            res *= spider1(grid) * spider2(grid) * spider3(grid)

        return Field(res, grid)

    if with_spiders and return_segments:
        # Use function to return the lambda, to avoid incorrect binding of variables
        def segment_with_spider(segment):
            return lambda grid: segment(grid) * spider1(grid) * spider2(grid) * spider3(grid)

        segments = [segment_with_spider(s) for s in segments]

    if return_header:
        if return_segments:
            return func, aperture_header, segments
        else:
            return func, aperture_header
    elif return_segments:
        return func, segments
    else:
        return func

def make_luvoir_a_lyot_stop(
        normalized=False, with_spiders=False, spider_oversize=1,
        inner_diameter_fraction=0.2, outer_diameter_fraction=0.9, return_header=False):
    '''Make a LUVOIR-A Lyot stop for the APLC coronagraph.

    Parameters
    ----------
    normalized : boolean
        If this is True, the pupil diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 15.0 meters.
    with_spiders : boolean
        Include the secondary mirror support structure in the aperture.
    inner_diameter_fraction : scalar
        The fractional size of the circular central obstruction as fraction of the pupil diameter.
    outer_diameter_fraction : scalar
        The fractional size of the circular outer edge as fraction of the pupil diameter.
    spider_oversize : scalar
        The factor by which to oversize the spiders compared to the LUVOIR-A aperture spiders.
    return_header : boolean
        If this is True, a header will be returned giving all important values for the
        created aperture for reference.

    Returns
    -------
    lyot_stop : Field generator
        A field generator for the Lyot stop.
    header : dict
        A dictionary containing all important values for the created aperture. Only returned
        if `return_header` is True.
    '''
    pupil_diameter = 15.0  # m actual circumscribed diameter, used for lam/D calculations other measurements normalized by this diameter
    spider_width = 0.150  # m actual strut size
    lower_spider_angle = 12.7  # deg angle at which lower spiders are offset from vertical
    spid_start = 0.30657  # m spider starting point offset from center of aperture

    outer_D = pupil_diameter * outer_diameter_fraction
    inner_D = pupil_diameter * inner_diameter_fraction
    pad_spid_width = spider_width * spider_oversize

    lyot_reference_diameter = pupil_diameter

    ls_header = {
        'TELESCOP': 'LUVOIR A',
        'D_CIRC': pupil_diameter,
        'D_INSC': 13.5,
        'LS_ID': inner_diameter_fraction,
        'LS_OD': outer_diameter_fraction,
        'LS_REF_D': lyot_reference_diameter,
        'NORM': normalized,
        'STRUT_ST': spid_start
    }

    if with_spiders:
        ls_header['STRUT_W']  = spider_width
        ls_header['STRUT_AN'] = lower_spider_angle
        ls_header['STRUT_P']  = spider_oversize

    if normalized:
        outer_D /= pupil_diameter
        inner_D /= pupil_diameter
        pad_spid_width /= pupil_diameter
        spid_start /= pupil_diameter

    outer_diameter = make_circular_aperture(outer_D)
    central_obscuration = make_circular_aperture(inner_D)

    if with_spiders:
        spider1 = make_spider_infinite([0, 0], 90, pad_spid_width)
        spider2 = make_spider_infinite([spid_start, 0], 270 - lower_spider_angle, pad_spid_width)
        spider3 = make_spider_infinite([-spid_start, 0], 270 + lower_spider_angle, pad_spid_width)

    def aper(grid):
        result = outer_diameter(grid) - central_obscuration(grid)

        if with_spiders:
            result *= spider1(grid) * spider2(grid) * spider3(grid)

        return result

    if return_header:
        return aper, ls_header

    return aper

def make_luvoir_b_aperture(
        normalized=False, with_segment_gaps=True, gap_padding=1, segment_transmissions=1,
        return_header=False, return_segments=False):
    '''Make the LUVOIR B aperture.

    This aperture is based on the LUVOIR final report 2019.
    Segment gaps can be included or excluded, and the transmission for each
    of the segments can also be changed. Segments can be returned as well.

    Parameters
    ----------
    normalized : boolean
        If this is True, the pupil diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 8.0 meters.
    with_segment_gaps : boolean
        Include the gaps between individual segments in the aperture.
    gap_padding : scalar
        Arbitrary padding of gap size to represent gaps on smaller arrays - this effectively
        makes the gaps larger and the segments smaller to preserve the same segment pitch.
    segment_transmissions : scalar or array_like
        The transmission for each of the segments. If this is a scalar, this transmission
        will be used for all segments.
    return_header : boolean
        If this is True, a header will be returned giving all important values for the
        created aperture for reference.
    return_segments : boolean
        If this is True, the segments will also be returned as a ModeBasis.

    Returns
    -------
    aperture : Field generator
        The LUVOIR B aperture.
    aperture_header : dict
        A dictionary containing all quantities used when making this aperture. Only returned if
        `return_header` is True.
    segments : list of Field generators
        The segments. Only returned when `return_segments` is True.
    '''
    pupil_diameter = 8.0  # m actual circumscribed diameter, used for lam/D calculations other measurements normalized by this diameter
    pupil_inscribed = 6.7  # m actual inscribed diameter
    actual_segment_flat_diameter = 0.955  # m actual segment flat-to-flat diameter
    actual_segment_gap = 0.006  # m actual gap size between segments

    # padding out the segmentation gaps so they are visible and not sub-pixel
    segment_gap = actual_segment_gap * gap_padding
    if not with_segment_gaps:
        segment_gap = 0

    segment_flat_diameter = actual_segment_flat_diameter - (segment_gap - actual_segment_gap)
    segment_circum_diameter = 2 / np.sqrt(3) * segment_flat_diameter

    num_rings = 4  # number of full rings of hexagons around central segment

    if not with_segment_gaps:
        segment_gap = 0

    aperture_header = {
        'TELESCOP': 'LUVOIR B',
        'D_CIRC': pupil_diameter,
        'D_INSC': pupil_inscribed,
        'SEG_F2F_D': actual_segment_flat_diameter,
        'SEG_GAP': actual_segment_gap,
        'NORM': normalized,
        'SEG_TRAN': segment_transmissions,
        'GAP_PAD': gap_padding,
        'PROV': 'LUVOIR final report 2019'
    }

    if normalized:
        segment_circum_diameter /= pupil_diameter
        actual_segment_flat_diameter /= pupil_diameter
        actual_segment_gap /= pupil_diameter
        pupil_diameter = 1.0

    segment_positions = make_hexagonal_grid(actual_segment_flat_diameter + actual_segment_gap, num_rings)

    # Clipping the "corner" segments of the outermost rings
    segment_positions = segment_positions.subset(make_circular_aperture(pupil_diameter * 0.9))

    segment = make_hexagonal_aperture(segment_circum_diameter, np.pi / 2)

    segmented_aperture = make_segmented_aperture(segment, segment_positions, segment_transmissions, return_segments=return_segments)

    if return_segments:
        segmented_aperture, segments = segmented_aperture

    def func(grid):
        res = segmented_aperture(grid)

        return Field(res, grid)

    if return_header:
        if return_segments:
            return func, aperture_header, segments
        else:
            return func, aperture_header
    elif return_segments:
        return func, segments
    else:
        return func

def make_hicat_aperture(normalized=False, with_spiders=True, with_segment_gaps=True, return_header=False, return_segments=False):
    '''Make the HiCAT P3 apodizer mask

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 15.0 meters.
    with_spiders : boolean
        Include the secondary mirror support structure in the aperture.
    with_segment_gaps : boolean
        Include the gaps between individual segments in the aperture.
    return_header : boolean
        If this is True, a header will be returned giving all important values for the
        created aperture for reference.
    return_segments : boolean
        If this is True, the segments will also be returned as a list of Field generators.

    Returns
    -------
    aperture : Field generator
        The HiCAT aperture.
    header : dict
        A dictionary containing all important values for the created aperture. Only returned
        if `return_header` is True.
    segments : list of Field generators
        The segments. Only returned when `return_segments` is True.
    '''
    gamma_21 = 0.423
    gamma_31 = 1.008

    # P2 - Iris AO
    p2_irisao_segment_size = 1.4e-3  # m (note: point to point)
    p2_irisao_segment_side_length = p2_irisao_segment_size / 2
    p2_irisao_segment_gap_size = 12e-6  # m

    p2_irisao_distance_between_segments = p2_irisao_segment_side_length * np.sqrt(3)
    p2_irisao_segment_circumdiameter = (2 * p2_irisao_segment_side_length) - (2 / np.sqrt(3)) * p2_irisao_segment_gap_size

    # P3 - Apodizer
    # Contour
    p3_apodizer_size = 19.725e-3  # m

    # Gap
    p3_apodizer_mask_gap_size = 0.090e-3  # m
    p3_irisao_segment_gap_size = p2_irisao_segment_gap_size * gamma_31 / gamma_21
    apodizer_mask_gap_oversize_factor_wrt_irisao = p3_apodizer_mask_gap_size / p3_irisao_segment_gap_size

    # Central segment
    p3_apodizer_mask_central_segment_size = 3.950e-3  # m

    # Spiders
    p3_apodizer_mask_spiders_thickness = 0.350e-3  # m

    header = {
        'TELESCOP': 'HiCAT',
        'P3_APOD': p3_apodizer_size,
        'P3_CENT_SEG': p3_apodizer_mask_central_segment_size,
        'P3_GAP': p3_apodizer_mask_gap_size,
        'P3_GAP_OVER': apodizer_mask_gap_oversize_factor_wrt_irisao,
        'P3_STRUT': p3_apodizer_mask_spiders_thickness,
        'PROV': 'HiCAT spreadsheet'
    }

    p3_irisao_segment_circumdiameter = p2_irisao_segment_circumdiameter * gamma_31 / gamma_21
    p3_irisao_distance_between_segments = p2_irisao_distance_between_segments * gamma_31 / gamma_21
    p3_apodizer_segment_circumdiameter = p3_irisao_segment_circumdiameter + (-p3_apodizer_mask_gap_size + p3_irisao_segment_gap_size) * (2 / np.sqrt(3))

    if normalized:
        p3_apodizer_segment_circumdiameter /= p3_apodizer_size
        p3_irisao_distance_between_segments /= p3_apodizer_size
        p3_apodizer_mask_central_segment_size /= p3_apodizer_size
        p3_apodizer_mask_spiders_thickness /= p3_apodizer_size
        p3_apodizer_size = 1

    segment = make_hexagonal_aperture(p3_apodizer_segment_circumdiameter, np.pi / 2)
    segment_positions = make_hexagonal_grid(p3_irisao_distance_between_segments, 3, False)
    segmentation = make_segmented_aperture(segment, segment_positions)

    segment = make_hexagonal_aperture(p3_apodizer_size / 7 / np.sqrt(3) * 2, np.pi / 2)
    distance_between_segments = p3_apodizer_size / 7
    segment_positions = make_hexagonal_grid(distance_between_segments, 3)
    contour = make_segmented_aperture(segment, segment_positions, return_segments=return_segments)

    if return_segments:
        contour, segments = contour

    central_segment = make_hexagonal_aperture(p3_apodizer_mask_central_segment_size, np.pi / 2)

    if with_spiders:
        spider1 = make_spider_infinite([0, 0], 60, p3_apodizer_mask_spiders_thickness)
        spider2 = make_spider_infinite([0, 0], 120, p3_apodizer_mask_spiders_thickness)
        spider3 = make_spider_infinite([0, 0], -60, p3_apodizer_mask_spiders_thickness)
        spider4 = make_spider_infinite([0, 0], -120, p3_apodizer_mask_spiders_thickness)

    def func(grid):
        res = contour(grid) - central_segment(grid)

        if with_segment_gaps:
            res *= segmentation(grid)

        if with_spiders:
            res *= spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid)

        return Field(res, grid)

    if return_segments:
        def segment_with_obstructions(seg):
            def func2(grid):
                return Field(func(grid) * seg(grid), grid)

            return func2

        segments = [segment_with_obstructions(s) for s in segments]

    if return_header:
        if return_segments:
            return func, header, segments
        else:
            return func, header
    elif return_segments:
        return func, segments
    else:
        return func

def make_hicat_lyot_stop(normalized=False, with_spiders=True, inner_diameter_fraction=0.2, outer_diameter_fraction=0.9, return_header=False):
    '''Make the HiCAT Lyot stop.

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 15.0 meters.
    with_spiders : boolean
        Include the secondary mirror support structure in the aperture.
    inner_diameter_fraction : scalar
        The fractional size of the circular central obstruction as fraction of the pupil diameter.
    outer_diameter_fraction : scalar
        The fractional size of the circular outer edge as fraction of the pupil diameter.
    return_header : boolean
        If this is True, a header will be returned giving all important values for the
        created aperture for reference.

    Returns
    -------
    lyot_stop : Field generator
        A field generator for the Lyot stop.
    header : dict
        A dictionary containing all important values for the created aperture. Only returned
        if `return_header` is True.
    '''
    gamma_31 = 1.008
    gamma_51 = 0.979

    p3_apodizer_mask_central_segment_size = 3.950e-3  # m
    p3_apodizer_size = 19.725e-3  # m
    p5_apodizer_size = p3_apodizer_size * gamma_51 / gamma_31

    p5_lyot_stop_size = outer_diameter_fraction * p5_apodizer_size  # m

    # Central segment
    p5_lyot_stop_mask_central_segment_size = inner_diameter_fraction * p5_apodizer_size  # m

    # Spiders
    p5_lyot_stop_mask_spiders_thickness = 0.700e-3  # m

    if normalized:
        p5_lyot_stop_size /= p5_apodizer_size
        p5_lyot_stop_mask_central_segment_size /= p5_apodizer_size
        p5_lyot_stop_mask_spiders_thickness /= p5_apodizer_size

    central_obscuration = make_circular_aperture(p5_lyot_stop_mask_central_segment_size)
    outer_diameter = make_circular_aperture(p5_lyot_stop_size)

    header = {
        'TELESCOP': 'HiCAT',
        'P3_APOD': p3_apodizer_size,
        'P3_CENT_SEG': p3_apodizer_mask_central_segment_size,
        'LS_CENT': p5_lyot_stop_mask_central_segment_size,
        'LS_SIZE': p5_lyot_stop_size,
        'P5_STRUT': p5_lyot_stop_mask_spiders_thickness
    }

    if with_spiders:
        spider1 = make_spider_infinite([0, 0], 60, p5_lyot_stop_mask_spiders_thickness)
        spider2 = make_spider_infinite([0, 0], 120, p5_lyot_stop_mask_spiders_thickness)
        spider3 = make_spider_infinite([0, 0], -60, p5_lyot_stop_mask_spiders_thickness)
        spider4 = make_spider_infinite([0, 0], -120, p5_lyot_stop_mask_spiders_thickness)

    def func(grid):
        res = outer_diameter(grid) - central_obscuration(grid)

        if with_spiders:
            res *= spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid)

        return Field(res, grid)

    if return_header:
        return func, header
    else:
        return func

def make_elt_aperture(normalized=False, with_spiders=True, segment_transmissions=1, return_segments=False):
    '''Make the European Extremely Large Telescope aperture.

    This aperture is based on Figure 3.66 that describes the pupil in the E-ELT Construction Proposal:
        https://www.eso.org/sci/facilities/eelt/docs/index.html .

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 39.14634 meters.
    with_spiders : boolean
        If this is False, the spiders will be left out. Default: True.
    segment_transmissions : scalar or array_like
        The transmission for each of the segments. If this is a scalar, this transmission
        will be used for all segments.
    return_segments : boolean
        If this is True, the segments will also be returned as a list of Field generators.

    Returns
    -------
    Field generator
        The E-ELT aperture.
    elt_segments : list of Field generators
        The segments. Only returned when `return_segments` is True.
    '''
    elt_outer_diameter = 39.14634
    spider_width = 0.4
    segment_size = 1.45
    segment_gap = 0.004
    inner_diameter = 9.4136
    outer_diameter = 39.14634

    if normalized:
        segment_size /= elt_outer_diameter
        segment_gap /= elt_outer_diameter
        inner_diameter /= elt_outer_diameter
        outer_diameter /= elt_outer_diameter
        spider_width /= elt_outer_diameter

    segment_positions = make_hexagonal_grid(segment_size * np.sqrt(3) / 2 + segment_gap, 17, pointy_top=False)

    # remove the inner segments
    central_obscuration_mask = (1 - make_hexagonal_aperture(inner_diameter * 2 / np.sqrt(3))(segment_positions)) > 0
    segment_positions = segment_positions.subset(central_obscuration_mask)

    # remove the pointy tops for a more circular aperture
    edge_mask_top = abs(segment_positions.y) < ((outer_diameter / 2) * 0.99)
    edge_mask_positive_30 = abs(np.cos(np.pi / 6) * segment_positions.x + np.sin(np.pi / 6) * segment_positions.y) < ((outer_diameter / 2) * 0.99)
    edge_mask_negative_30 = abs(np.cos(np.pi / 6) * segment_positions.x - np.sin(np.pi / 6) * segment_positions.y) < ((outer_diameter / 2) * 0.99)
    all_edge_masks = edge_mask_top * edge_mask_positive_30 * edge_mask_negative_30 > 0

    segment_positions = segment_positions.subset(all_edge_masks)

    segment_shape = make_hexagonal_aperture(segment_size, angle=np.pi / 2)

    if return_segments:
        elt_aperture_function, elt_segments = make_segmented_aperture(segment_shape, segment_positions, segment_transmissions, return_segments=return_segments)
    else:
        elt_aperture_function = make_segmented_aperture(segment_shape, segment_positions, segment_transmissions)

    spiders = [make_spider_infinite([0, 0], 60 * i + 30, spider_width) for i in range(6)]

    def elt_aperture_with_spiders(grid):
        aperture = elt_aperture_function(grid)

        if with_spiders:
            for spider in spiders:
                aperture *= spider(grid)

        return aperture

    if with_spiders and return_segments:
        # Use function to return the lambda, to avoid incorrect binding of variables
        def spider_func(grid):
            spider_aperture = grid.ones()
            for spider in spiders:
                spider_aperture *= spider(grid)
            return spider_aperture

        def segment_with_spider(segment):
            return lambda grid: segment(grid) * spider_func(grid)

        elt_segments = [segment_with_spider(s) for s in elt_segments]

    if return_segments:
        return elt_aperture_with_spiders, elt_segments
    else:
        return elt_aperture_with_spiders

def make_gmt_aperture(normalized=False, with_spiders=True, return_segments=False):
    '''Make the Giant Magellan Telescope aperture.

    The primary mirror parameters come from the GMT Observatory Architecture Documentation (GMT-REQ-03215, Rev. F):
    https://www.gmto.org/resources/slpdr/ . Small corrections have been applied to match to the actual pupil from internal GMTO files.

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 25.448 meters.
    with_spiders : boolean
        If this is False, the spiders will be left out. Default: True.
    return_segments : boolean
        If this is True, the segments will also be returned as a list of Field generators.

    Returns
    -------
    Field generator
        The GMT aperture.
    elt_segments : list of Field generators
        The segments. Only returned when `return_segments` is True.
    '''
    gmt_outer_diameter = 25.448
    segment_size = 8.365 - 0.072
    off_axis_segment_size = 8.365 - 0.015

    # The spider truss to hold the secondary
    spider_1_width = 0.119
    spider_2_width = 0.115
    radius_spider_1 = 2.386
    radius_spider_2 = 2.409
    offset_spider_2 = -0.05
    truss_size = 4.93

    # 0.359 mm from the off-axis segment to the on-axis segment
    segment_gap = 0.359 + 0.088
    off_axis_tilt = np.deg2rad(13.522)
    central_hole_size = 3.495 / segment_size
    segment_distance = segment_size / 2 + off_axis_segment_size / 2 * np.cos(off_axis_tilt) + segment_gap

    if normalized:
        segment_size /= gmt_outer_diameter
        off_axis_segment_size /= gmt_outer_diameter
        segment_gap /= gmt_outer_diameter
        segment_distance /= gmt_outer_diameter
        truss_size /= gmt_outer_diameter
        spider_1_width /= gmt_outer_diameter
        spider_2_width /= gmt_outer_diameter
        radius_spider_1 /= gmt_outer_diameter
        radius_spider_2 /= gmt_outer_diameter
        offset_spider_2 /= gmt_outer_diameter

    def make_diverging_spider(position, start_width, divergence, orientation):
        def func(grid):
            y = grid.shifted(position).rotated(orientation).y
            x = grid.shifted(position).rotated(orientation).x
            return Field(abs(y) < (np.sin(divergence) * abs(x) + start_width / 2) * (x >= 0), grid)
        return func

    def make_central_gmt_segment(grid):
        center_segment = make_obstructed_circular_aperture(segment_size, central_hole_size)(grid)
        if with_spiders:
            spider_attachement_mask = 1 - make_regular_polygon_aperture(3, truss_size, angle=np.pi, center=None)(grid)

            spider_mask = grid.ones()
            for i in range(3):
                offset_angle = 2 * np.pi / 3 * i
                spider_1_start = radius_spider_1 * np.array([np.sin(offset_angle), np.cos(offset_angle)])
                spider_mask *= 1 - make_diverging_spider(spider_1_start, spider_1_width, np.deg2rad(1.16), np.deg2rad(9.83) + offset_angle)(grid)
                spider_mask *= 1 - make_diverging_spider(spider_1_start, spider_1_width, np.deg2rad(1.16), np.deg2rad(180.0 - 9.83) + offset_angle)(grid)

                spider_2_start = radius_spider_2 * np.array([np.sin(offset_angle), np.cos(offset_angle)]) + offset_spider_2 * np.array([np.cos(offset_angle), -np.sin(offset_angle)])
                spider_mask *= 1 - make_diverging_spider(spider_2_start, spider_2_width, np.deg2rad(0.0), np.deg2rad(-11.2) + offset_angle)(grid)

            return center_segment * spider_mask * spider_attachement_mask
        else:
            return center_segment

    segment_functions = [make_central_gmt_segment]
    for i in range(6):
        rotation_angle = np.pi / 3 * i
        xc = segment_distance * np.cos(rotation_angle)
        yc = segment_distance * np.sin(rotation_angle)

        aperture = make_elliptical_aperture([off_axis_segment_size * np.cos(off_axis_tilt), off_axis_segment_size], center=[xc, yc], angle=-rotation_angle)
        segment_functions.append(aperture)

    # Center segment obscurations
    def make_aperture(grid):
        aperture = grid.zeros()
        for segment in segment_functions:
            aperture += segment(grid)
        return aperture

    if return_segments:
        return make_aperture, segment_functions
    else:
        return make_aperture

def make_tmt_aperture(normalized=False, with_spiders=True, segment_transmissions=1, return_segments=False):
    '''Make the Thirty-Meter Telescope aperture.

    The aperture is based on the description from https://www.tmt.org/page/optics. The size of
    the secondary and the spiders were derived from Figure 5 of [JensenClem2021]_

    .. [JensenClem2021] Rebecca Jensen-Clem et al. "The Planetary Systems Imager Adaptive Optics System:
        An Initial Optical Design and Performance Analysis Tools for the PSI-Red AO System" SPIE Vol. 11823 (2021)

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 30.0 meters.
    with_spiders : boolean
        If this is False, the spiders will be left out. Default: True.
    segment_transmissions : scalar or array_like
        The transmission for each of the segments. If this is a scalar, this transmission
        will be used for all segments.
    return_segments : boolean
        If this is True, the segments will also be returned as a list of Field generators.

    Returns
    -------
    Field generator
        The TMT aperture.
    tmt_segments : list of Field generators
        The segments. Only returned when `return_segments` is True.
    '''
    tmt_outer_diameter = 30.0     # meter
    spider_width = 0.22     # meter
    segment_size = 1.44     # meter
    segment_gap = 0.0025     # meter
    inner_diameter = 2.5 * segment_size
    central_obscuration = 3.636     # meter

    if normalized:
        spider_width /= tmt_outer_diameter
        segment_size /= tmt_outer_diameter
        segment_gap /= tmt_outer_diameter
        inner_diameter /= tmt_outer_diameter
        central_obscuration /= tmt_outer_diameter
        tmt_outer_diameter = 1.0

    segment_positions = make_hexagonal_grid(segment_size * np.sqrt(3) / 2 + segment_gap, 13, pointy_top=False)

    # remove the first ring and the central segment
    missing_segments_mask = (1 - make_hexagonal_aperture(inner_diameter * 2 / np.sqrt(3))(segment_positions)) > 0
    segment_positions = segment_positions.subset(missing_segments_mask)

    # Clip the outer segments
    inscribed_circle = make_circular_aperture(0.98 * tmt_outer_diameter)(segment_positions) > 0
    segment_positions = segment_positions.subset(inscribed_circle)

    segment_shape = make_hexagonal_aperture(segment_size, angle=np.pi / 2)

    if return_segments:
        tmt_aperture_function, tmt_segments = make_segmented_aperture(segment_shape, segment_positions, segment_transmissions, return_segments=return_segments)
    else:
        tmt_aperture_function = make_segmented_aperture(segment_shape, segment_positions, segment_transmissions)

    spiders = [make_spider_infinite([0, 0], 60 * i + 30, spider_width) for i in range(6)]

    def tmt_aperture_with_spiders(grid):
        aperture = tmt_aperture_function(grid) * (1 - make_circular_aperture(central_obscuration)(grid))

        if with_spiders:
            for spider in spiders:
                aperture *= spider(grid)

        return aperture

    if with_spiders and return_segments:
        # Use function to return the lambda, to avoid incorrect binding of variables
        def spider_func(grid):
            spider_aperture = grid.ones()
            for spider in spiders:
                spider_aperture *= spider(grid)
            return spider_aperture

        def segment_with_spider(segment):
            return lambda grid: segment(grid) * spider_func(grid) * (1 - make_circular_aperture(central_obscuration)(grid))

        tmt_segments = [segment_with_spider(s) for s in tmt_segments]
    elif not with_spiders and return_segments:
        def segment_with_central_obscuration(segment):
            return lambda grid: segment(grid) * (1 - make_circular_aperture(central_obscuration)(grid))

        tmt_segments = [segment_with_central_obscuration(s) for s in tmt_segments]

    if return_segments:
        return tmt_aperture_with_spiders, tmt_segments
    else:
        return tmt_aperture_with_spiders

def make_habex_aperture(normalized=False):
    '''Make the HabEx aperture.

    This aperture is based on the HabEx final report 2019. As the HabEx is unobscured,
    there is no option to exclude spiders.

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 4.0 meters.

    Returns
    -------
    Field generator
        The HabEx telescope aperture.
    '''
    pupil_diameter = 4.0  # meter

    if normalized:
        pupil_diameter = 1

    return make_circular_aperture(pupil_diameter)

def make_hst_aperture(normalized=False, with_spiders=True, with_pads=True):
    '''Make the Hubble Space Telescope aperture.

    This function uses values from TinyTim [Krist2011]_.

    .. [Krist2011] Krist, Hook and Stoehr, "20 years of Hubble Space Telescope optical modelling using Tiny Tim" Proc. SPIE 8127 (2011).

    .. note::
        This aperture only includes the primary and secondary masks, not any masks
        internal to the instruments.

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 2.4 meters.
    with_spiders: boolean
        If this is False, the spiders will be left out.

    Returns
    -------
    Field generator
        The Hubble Space Telescope aperture.
    '''
    pupil_diameter = 2.4  # meter

    # All these values are relative to the pupil diameter.
    secondary_obscuration_ratio = 0.330
    spider_width = 0.022 / 2

    pad_v3 = np.array([0.8921, -0.4615, -0.4564]) / 2
    pad_v2 = np.array([0.0000, 0.7555, -0.7606]) / 2
    pad_radii = np.array([0.065, 0.065, 0.065]) / 2

    if normalized:
        pupil_diameter = 1
    else:
        spider_width *= pupil_diameter

        pad_v3 *= pupil_diameter
        pad_v2 *= pupil_diameter
        pad_radii *= pupil_diameter

    num_spiders = 4 if with_spiders else 0

    ota = make_obstructed_circular_aperture(pupil_diameter, secondary_obscuration_ratio, num_spiders, spider_width)

    if not with_pads:
        return ota

    pads = [make_obstruction(make_circular_aperture(2 * r, [-v2, v3])) for v3, v2, r in zip(pad_v3, pad_v2, pad_radii)]

    def func(grid):
        return ota(grid) * pads[0](grid) * pads[1](grid) * pads[2](grid)

    return func

def make_jwst_aperture(normalized=False, with_spiders=True, return_segments=False):
    '''Make James Webb Space Telescope (JWST) aperture.

    This function uses coordinates from [WebbPSF]_ v1.0.0. WebbPSF itself got the
    coordinate values from an internal spreadsheet by Paul Lightsey "2010.03.16
    Transmission X Area Budget.xls", which was based in turn on Ball Aerospace
    drawing # 2220169 Rev B and the OTE Cryogenic Optics Interface Control Document,
    Ball Aerospace doc # C327693.

    .. [WebbPSF] The James Webb Space Telescope PSF simulation tool.
        (https://github.com/spacetelescope/webbpsf)

    .. note::
        This function uses the pre-launch aperture for JWST and does not include
        any optical difference maps, filter transmission curves or any internal
        masks in any of the instruments. Use WebbPSF for more complete simulations.

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 6.603464 meters.
    with_spiders : boolean
        If this is False, the spiders will be left out. Default: True.
    return_segments : boolean
        If this is True, the segments will also be returned as a list of Field generators.

    Returns
    -------
    aperture : Field generator
        The JWST aperture.
    segments : list of Field generators
        The segments. Only returned when `return_segments` is True.
    '''
    pupil_diameter = 6.603464  # meter

    segment_parameters = {
        'A1-1': np.array([
            [-0.38101, 0.667604],
            [-0.758826, 1.321999],
            [-0.38101, 1.976407],
            [0.38101, 1.976407],
            [0.758826, 1.321999],
            [0.38101, 0.667604]]),
        'A2-2': np.array([
            [0.38765702, 0.66376634],
            [0.76547172, 1.31816209],
            [1.52111367, 1.31816784],
            [1.90212367, 0.65823916],
            [1.52429772, 0.00383691],
            [0.76866702, 0.00383766]]),
        'A3-3': np.array([
            [0.76866702, -0.00383766],
            [1.52429772, -0.00383691],
            [1.90212367, -0.65823916],
            [1.52111367, -1.31816784],
            [0.76547172, -1.31816209],
            [0.38765702, -0.66376634]]),
        'A4-4': np.array([
            [0.38101, -0.667604],
            [0.758826, -1.321999],
            [0.38101, -1.976407],
            [-0.38101, -1.976407],
            [-0.758826, -1.321999],
            [-0.38101, -0.667604]]),
        'A5-5': np.array([
            [-0.38765702, -0.66376634],
            [-0.76547172, -1.31816209],
            [-1.52111367, -1.31816784],
            [-1.90212367, -0.65823916],
            [-1.52429772, -0.00383691],
            [-0.76866702, -0.00383766]]),
        'A6-6': np.array([
            [-0.76866702, 0.00383766],
            [-1.52429772, 0.00383691],
            [-1.90212367, 0.65823916],
            [-1.52111367, 1.31816784],
            [-0.76547172, 1.31816209],
            [-0.38765702, 0.66376634]]),
        'B1-7': np.array([
            [0.38101, 3.279674],
            [0.758826, 2.631791],
            [0.38101, 1.98402],
            [-0.38101, 1.98402],
            [-0.758826, 2.631791],
            [-0.38101, 3.279674]]),
        'B2-9': np.array([
            [3.030786, 1.30987266],
            [2.65861086, 0.65873291],
            [1.90871672, 0.66204566],
            [1.52770672, 1.32197434],
            [1.89978486, 1.97305809],
            [2.649776, 1.96980134]]),
        'B3-11': np.array([
            [2.649776, -1.96980134],
            [1.89978486, -1.97305809],
            [1.52770672, -1.32197434],
            [1.90871672, -0.66204566],
            [2.65861086, -0.65873291],
            [3.030786, -1.30987266]]),
        'B4-13': np.array([
            [-0.38101, -3.279674],
            [-0.758826, -2.631791],
            [-0.38101, -1.98402],
            [0.38101, -1.98402],
            [0.758826, -2.631791],
            [0.38101, -3.279674]]),
        'B5-15': np.array([
            [-3.030786, -1.30987266],
            [-2.65861086, -0.65873291],
            [-1.90871672, -0.66204566],
            [-1.52770672, -1.32197434],
            [-1.89978486, -1.97305809],
            [-2.649776, -1.96980134]]),
        'B6-17': np.array([
            [-2.649776, 1.96980134],
            [-1.89978486, 1.97305809],
            [-1.52770672, 1.32197434],
            [-1.90871672, 0.66204566],
            [-2.65861086, 0.65873291],
            [-3.030786, 1.30987266]]),
        'C1-8': np.array([
            [0.765201, 2.627516],
            [1.517956, 2.629178],
            [1.892896, 1.976441],
            [1.521076, 1.325812],
            [0.765454, 1.325807],
            [0.387649, 1.980196]]),
        'C2-10': np.array([
            [2.6580961, 0.651074495],
            [3.03591294, 5.42172989e-07],
            [2.65809612, -0.651075523],
            [1.90872487, -0.654384457],
            [1.53090954, 8.90571587e-07],
            [1.90872454, .654384118]]),
        'C3-12': np.array([
            [1.8928951, -1.97644151],
            [1.51795694, -2.62917746],
            [0.76520012, -2.62751652],
            [0.38764887, -1.98019646],
            [0.76545554, -1.32580611],
            [1.52107554, -1.32581188]]),
        'C4-14': np.array([
            [-0.765201, -2.627516],
            [-1.517956, -2.629178],
            [-1.892896, -1.976441],
            [-1.521076, -1.325812],
            [-0.765454, -1.325807],
            [-0.387649, -1.980196]]),
        'C5-16': np.array([
            [-2.6580961, -.651074495],
            [-3.03591294, -5.42172990e-07],
            [-2.65809612, .651075523],
            [-1.90872487, .654384457],
            [-1.53090954, -8.90571587e-07],
            [-1.90872454, -.654384118]]),
        'C6-18': np.array([
            [-1.8928951, 1.97644151],
            [-1.51795694, 2.62917746],
            [-0.76520012, 2.62751652],
            [-0.38764887, 1.98019646],
            [-0.76545554, 1.32580611],
            [-1.52107554, 1.32581188]]),
    }

    strut_parameters = {
        "strut1": np.array([
            [-0.05301375, -0.0306075],
            [1.59698625, -2.88849133],
            [1.70301375, -2.82727633],
            [0.05301375, 0.0306075],
            [-0.05301375, -0.0306075]]),
        "strut2": np.array([
            [-0.05301375, 0.0306075],
            [-1.70301375, -2.82727633],
            [-1.59698625, -2.88849133],
            [0.05301375, -0.0306075],
            [-0.05301375, 0.0306075]]),
        "strut3": np.array([
            [5.94350000e-02, -1.45573765e-17],
            [5.94350000e-02, 3.30000000e+00],
            [-5.94350000e-02, 3.30000000e+00],
            [-5.94350000e-02, 1.45573765e-17],
            [5.94350000e-02, -1.45573765e-17]]),
        "strut3_bumps": np.array([
            [0.059435, 0.666],
            [0.059435, 2.14627],
            [0.082595, 2.14627],
            [0.082595, 2.3645],
            [0.059435, 2.3645],
            [0.059435, 2.48335],
            [0.069795, 2.48335],
            [0.069795, 2.54445],
            [0.059435, 2.54445],
            [0.059435, 3.279674],
            [-0.059435, 3.279674],
            [-0.059435, 2.54445],
            [-0.069795, 2.54445],
            [-0.069795, 2.48335],
            [-0.059435, 2.48335],
            [-0.059435, 2.3645],
            [-0.082595, 2.3645],
            [-0.082595, 2.14627],
            [-0.059435, 2.14627],
            [-0.059435, 0.666]])
    }

    if normalized:
        for val in segment_parameters.values():
            val /= pupil_diameter

        for val in strut_parameters.values():
            val /= pupil_diameter

        pupil_diameter = 1

    segments = [make_irregular_polygon_aperture(vertices) for vertices in segment_parameters.values()]
    struts = [make_obstruction(make_irregular_polygon_aperture(vertices)) for vertices in strut_parameters.values()]

    def func(grid):
        res = sum((segment(grid) for segment in segments))

        if with_spiders:
            for strut in struts:
                res *= strut(grid)

        return res

    if not return_segments:
        return func

    if with_spiders:
        def segment_with_strut(segment):
            def func(grid):
                res = segment(grid)

                for strut in struts:
                    res *= strut(grid)

                return res
            return func

        segments = [segment_with_strut(segment) for segment in segments]

    return func, segments

def make_keck_aperture(normalized=False, with_spiders=True, with_segment_gaps=True, gap_padding=10, segment_transmissions=1, return_segments=False):
    """Make the Keck aperture.

    This code creates a Keck-like aperture matching values used in [vanKooten2022a]_ and [vanKooten2022b]_ as well as
    being verified by Keck personnel to match internal simulation efforts.

    .. [vanKooten2022a] Maaike van Kooten et al., "Predictive wavefront control on Keck II adaptive optics bench: on-sky coronagraphic results." JATIS 8 (2022): 029006
    .. [vanKooten2022b] Maaike van Kooten et al., "On-sky Reconstruction of Keck Primary Mirror Piston Offsets Using a Zernike Wavefront Sensor." The Astrophysical Journal 932 (2022): 2, 109.

    Parameters
    ----------
    normalized : boolean
        If this is True, the outer diameter will be scaled to 1. Otherwise, the
        diameter of the pupil will be 10.95 meters.
    with_spiders : boolean
        Include the secondary mirror support structure in the aperture.
    with_segment_gaps : boolean
        Include the gaps between individual segments in the aperture.
    gap_padding : scalar
        Arbitrary padding of gap size to represent gaps on smaller arrays - this effectively
        makes the gaps larger and the segments smaller to preserve the same segment pitch.
    segment_transmissions : scalar or array_like
        The transmission for each of the segments. If this is a scalar, this transmission
        will be used for all segments.
    return_segments : boolean
        If this is True, the segments will also be returned as a list of Field generators.

    Returns
    -------
    aperture : Field generator
        The Keck aperture.
    segments : list of Field generators
        The segments. Only returned when `return_segments` is True.
    """
    pupil_diameter = 10.95  # m actual circumscribed diameter
    actual_segment_flat_diameter = np.sqrt(3) / 2 * 1.8  # m actual segment flat-to-flat diameter
    central_obscuration_diameter = 2.6  # m
    actual_segment_gap = 0.003  # m actual gap size between segments
    spider_width = 2.6e-2  # m actual strut size
    num_rings = 3  # number of full rings of hexagons around central segment

    if normalized:
        actual_segment_flat_diameter /= pupil_diameter
        actual_segment_gap /= pupil_diameter
        spider_width /= pupil_diameter
        central_obscuration_diameter /= pupil_diameter
        pupil_diameter = 1.0

    # padding out the segmentation gaps so they are visible and not sub-pixel
    segment_gap = actual_segment_gap * gap_padding
    if not with_segment_gaps:
        segment_gap = 0

    segment_flat_diameter = actual_segment_flat_diameter - (segment_gap - actual_segment_gap)
    segment_circum_diameter = 2 / np.sqrt(3) * segment_flat_diameter

    segment_positions = make_hexagonal_grid(actual_segment_flat_diameter + actual_segment_gap, num_rings)

    segment = make_hexagonal_aperture(segment_circum_diameter, np.pi / 2)

    if with_spiders:
        spider1 = make_spider_infinite([0, 0], 0, spider_width)
        spider2 = make_spider_infinite([0, 0], 60, spider_width)
        spider3 = make_spider_infinite([0, 0], 120, spider_width)
        spider4 = make_spider_infinite([0, 0], 180, spider_width)
        spider5 = make_spider_infinite([0, 0], 240, spider_width)
        spider6 = make_spider_infinite([0, 0], 300, spider_width)

    segmented_aperture = make_segmented_aperture(segment, segment_positions, segment_transmissions, return_segments=return_segments)
    if return_segments:
        segmented_aperture, segments = segmented_aperture

    def func(grid):
        res = segmented_aperture(grid) * (1 - make_circular_aperture(central_obscuration_diameter)(grid))

        if with_spiders:
            res *= spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid) * spider5(grid) * spider6(grid)

        return Field(res, grid)

    if with_spiders and return_segments:
        # Use function to return the lambda, to avoid incorrect binding of variables
        def segment_with_spider(segment):
            return lambda grid: segment(grid) * spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid) * spider5(grid) * spider6(grid)

        segments = [segment_with_spider(s) for s in segments]

    if return_segments:
        def segment_with_central_obscuration(segment):
            return lambda grid: segment(grid) * (1 - make_circular_aperture(central_obscuration_diameter)(grid))

        segments = [segment_with_central_obscuration(segment) for segment in segments]

        return func, segments
    else:
        return func



import os
import yaml
from pathlib import Path

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

class _ConfigurationItem(object):
    def __init__(self, val):
        self._val = val

    def __getitem__(self, key):
        val = self._val[key]

        if isinstance(val, dict):
            return _ConfigurationItem(val)
        else:
            return val

    def __setitem__(self, key, value):
        self._val[key] = value

    def __contains__(self, key):
        return key in self._val

    def __getattr__(self, name):
        if name == '_val':
            return getattr(self, name)
        else:
            return self.__getitem__(name)

    def __setattr__(self, name, value):
        if name == '_val':
            super().__setattr__(name, value)
        else:
            self.__setitem__(name, value)

    def __str__(self):
        return str(self._val)

    def clear(self):
        self._val.clear()

    def update(self, b):
        for key in b.keys():
            if key in self:
                if isinstance(b[key], dict) and isinstance(self._val[key], dict):
                    self[key].update(b[key])
                else:
                    self[key] = b[key]
            else:
                self[key] = b[key]

class Configuration(object):
    '''A configuration object that describes the current configuration status of the package.
    '''
    def __init__(self):
        if Configuration._config is None:
            self.reset()

    _config = None

    def __getitem__(self, key):
        '''Get the value for the key.

        Parameters
        ----------
        key : string
            The configuration key.
        '''
        return Configuration._config[key]

    def __setitem__(self, key, value):
        '''Set the value for the key.

        Parameters
        ----------
        key : string
            The configuration key.
        value : anything
            The value to set this to.
        '''
        Configuration._config[key] = value

    __getattr__ = __getitem__
    __setattr__ = __setitem__

    def __str__(self):
        return str(Configuration._config)

    def reset(self):
        '''Reset the configuration to the default configuration.

        This default configuration consists of the default parameters in `hcipy/data/default_config.yaml`, which
        can be overridden by a configuration file in `~/.hcipy/hcipy_config.yaml`. This can in turn be overridden
        by a configuration file named `hcipy_config.yaml` located in the current working directory.
        '''
        Configuration._config = _ConfigurationItem({})

        default_config = files('hcipy.config').joinpath('default_config.yaml')
        user_config = Path(os.path.expanduser('~/.hcipy/hcipy_config.yaml'))
        current_working_directory = Path('./hcipy_config.yaml')

        paths = [default_config, user_config, current_working_directory]

        for path in paths:
            try:
                contents = path.read_text()
                new_config = yaml.safe_load(contents)

                self.update(new_config)
            except FileNotFoundError:
                # If a configuration file is not found, just ignore.
                pass

    def update(self, b):
        '''Update the configuration with the configuration `b`, as a dictionary.

        Parameters
        ----------
        b : dict
            A dictionary containing the values to update in the configuration.
        '''
        Configuration._config.update(b)



import numpy as np
from ..field import Field

from ..optics import GeometricPhaseElement

def generate_app_keller(wavefront, propagator, contrast, num_iterations, beta=0):
    """
    Accelerated Gerchberg-Saxton-like algorithm for APP design by
    Christoph Keller [Keller2016]_ and based on Douglas-Rachford operator splitting.
    The acceleration was inspired by the paper by Jim Fienup [Fienup1976]_. The
    acceleration can provide speed-ups of up to two orders of magnitude and
    produce better APPs.

    .. [Keller2016] Keller C.U., 2016, "Novel instrument concepts for
        characterizing directly imaged exoplanets", Proc. SPIE 9908,
        Ground-based and Airborne Instrumentation for Astronomy VI, 99089V
        doi: 10.1117/12.2232633; https://doi.org/10.1117/12.2232633

    .. [Fienup1976] J. R. Fienup, 1976, "Reconstruction of an object from the modulus
        of its Fourier transform," Opt. Lett. 3, 27-29

    Parameters
    ----------
    wavefont : Wavefront
        The input aperture as a wavefront; a phase can be provided as a
        starting point for the APP generation.
    propagator : Propagator
        The propagator from the pupil grid to the focal plane grid.
    contrast : array_like
        The required contrast in the focal plane; this is a float mask
        that is 1.0 everywhere except for the dark zone where it is the
        contrast such as 1e-5.
    num_iterations : int
        The maximum number of iterations.
    beta : scalar
        The acceleration parameter. The default is 0 (no acceleration).
        Good values for beta are typically between 0.3 and 0.9. Values larger
        than 1.0 will not work.

    Returns
    -------
    Wavefront
        The APP as a wavefront.

    Raises
    ------
    ValueError
        If beta is not between 0 and 1.
    """
    if beta < 0 or beta > 1:
        raise ValueError('Beta should be between 0 and 1.')

    # initialize APP with wavefront
    app = wavefront.copy()

    # define pupil aperture where amplitude is larger than 1e-6
    aperture = wavefront.amplitude > 1e-6

    # define dark zone as location where contrast requirement is < 1e-1
    dark_zone = contrast < 0.1

    old_image = None
    for i in range(num_iterations):
        # calculate image plane electric field
        image = propagator.forward(app)

        # stop iteration if contrast requirement is met
        if not np.any(image.intensity / image.intensity.max() > contrast):
            break

        # modify focal plane electic field using acceleration
        new_image = image.copy()
        if beta != 0 and old_image is not None:
            new_image.electric_field[dark_zone] = old_image.electric_field[dark_zone] * beta - new_image.electric_field[dark_zone] * (1 + beta)
        else:
            new_image.electric_field[dark_zone] = 0
        old_image = new_image.copy()

        # determine corresponding aperture electric field
        app.electric_field = propagator.backward(new_image).electric_field

        # enforce aperture support
        app.electric_field[~aperture] = 0

        # enforce unity transmission within aperture support
        app.electric_field[aperture] *= wavefront.amplitude[aperture] / app.amplitude[aperture]

    return app

def generate_app_por(wavefront, propagator, propagator_max, contrast, num_iterations=1):
    '''Optimize a one-sided APP using a globally optimal algorithm. This algorithm does not
    apply any symmetries for two-sided dark zones or circularly-symmetric pupils. This
    function requires that you have installed the Gurobi optimizer.

    Parameters
    ----------
    wavefont : Wavefront
        The input aperture as a wavefront.
    propagator : Propagator
        The propagator from `wavefront` to the focal plane grid.
    propagator_max : Propagator
        The propagator from `wavefront` to the point in the focal-plane that we want to maximize.
    contrast : array_like or scalar
        The required contrast in the focal plane.
    num_iterations : int
        The number of iterations for the Strehl ratio to converge.

    Returns
    -------
    Field
        The resultant electric field transmission function.
    '''
    import gurobipy as gp

    M = propagator.get_transformation_matrix_forward(wavefront.electric_field.grid, wavefront.wavelength)
    M_max = propagator_max.get_transformation_matrix_forward(wavefront.electric_field.grid, wavefront.wavelength)

    wf_0 = propagator.forward(wavefront).electric_field
    M /= wf_0[:, np.newaxis]

    M *= wavefront.electric_field
    M_max *= wavefront.electric_field

    M_max = np.sum(M_max, axis=0)

    M = np.array(np.bmat([[M.real, -M.imag], [M.imag, M.real]]))
    M_max = np.concatenate((M_max.real, -M_max.imag))

    m, n = M.shape
    contrast_requirement = np.ones(m) * np.sqrt(contrast)

    model = gp.Model('lp')
    model.Params.Threads = 4

    x = model.addVars(n, lb=-1, ub=1)

    obj = gp.quicksum((x[i] * M_max[i] for i in range(n)))
    model.setObjective(obj, gp.GRB.MAXIMIZE)

    for i in range(n // 2):
        r2 = x[i] * x[i] + x[i + n // 2] * x[i + n // 2]
        model.addQConstr(r2 <= 1)

    for i, ee in enumerate(M):
        e = gp.quicksum((x[i] * ee[i] for i in range(n)))
        model.addConstr(e <= contrast_requirement[i])
        model.addConstr(e >= -contrast_requirement[i])

    model.optimize()

    solution = np.array([x[i].x for i in range(n)])
    solution = solution[:n // 2] + 1j * solution[n // 2:]

    return Field(solution, wavefront.electric_field.grid)

def generate_app_doelman(wavefront, propagator, propagator_max, contrast, num_iterations):
    raise NotImplementedError()

class VectorApodizingPhasePlate(GeometricPhaseElement):
    '''The vector-Apodizing Phase Plate (vAPP).

    Parameters
    ----------
    phase_pattern : field or array_like
        The phase pattern generating the vAPP PSFs (radians).
    leakage : scalar
        The relative leakage strength (0 = no leakage, 1 = maximum leakage)
    retardance_offset : scalar
        The retardance offset from half wave in radians. This will result in leakage.
    '''
    def __init__(self, phase_pattern, leakage=None, retardance_offset=0):

        GeometricPhaseElement.__init__(self, phase_pattern, leakage, retardance_offset)



__all__ = [
    'MultiLayerAtmosphere',
    'AtmosphericLayer',
    'phase_covariance_von_karman',
    'phase_structure_function_von_karman',
    'power_spectral_density_von_karman',
    'Cn_squared_from_fried_parameter',
    'fried_parameter_from_Cn_squared',
    'seeing_to_fried_parameter',
    'fried_parameter_to_seeing',
    'FiniteAtmosphericLayer',
    'InfiniteAtmosphericLayer',
    'ModalAdaptiveOpticsLayer',
    'make_standard_atmospheric_layers',
    'make_las_campanas_atmospheric_layers',
    'make_keck_atmospheric_layers'
]

from .atmospheric_model import *
from .finite_atmospheric_layer import *
from .infinite_atmospheric_layer import *
from .modal_adaptive_optics_layer import *
from .standard_atmosphere import *



from ..field import Field, make_uniform_grid
import numpy as np
from scipy import sparse

def generate_convolution_matrix(grid, kernel):
    '''Create the matrix that applies a convolution with kernel.

    The created matrix is a sparse matrix.

    Parameters
    ----------
    grid : Grid
        The :class:`Grid` for which the convolution matrix will be created.
    kernel : Field or array_like
        The convolution kernel

    Returns
    -------
    array_like
        The matrix that applies the convolution.
    '''
    if hasattr(kernel, 'grid'):
        if np.allclose(kernel.grid.delta, grid.delta):
            num_x = kernel.grid.shape[1]
            num_y = kernel.grid.shape[0]
        else:
            raise ValueError("Kernel and grid are sampled with different grid spacings.")
    else:
        if kernel.ndim == 2:
            num_x = kernel.shape[1]
            num_y = kernel.shape[0]
            kernel = kernel.ravel()
        elif kernel.ndim == 1:
            raise NotImplementedError("Can not create a convolution kernel from a 1D array.")

    index_y, index_x = np.indices((num_y, num_x))
    offsets = ((index_x - num_x // 2) + (index_y - num_y // 2) * grid.shape[0]).ravel()
    convolution_matrix = sparse.diags(kernel, offsets, shape=(grid.size, grid.size))
    return convolution_matrix

def make_laplacian_matrix(grid):
    '''Make the Laplacian operator using the 5-point stencil approximation

    Parameters
    ----------
    grid : Grid
        The grid for which the derivative matrix is calculated.

    Returns
    -------
    array_like
        The convolution matrix.
    '''
    if grid.is_('cartesian') and grid.is_separated and grid.is_regular:
        num_x = 3
        num_y = 3
        kernel_grid = make_uniform_grid((num_x, num_y), (num_x * grid.delta[0], num_y * grid.delta[1]), has_center=True)
        kernel = kernel_grid.zeros().shaped
        kernel[1, 1] = 4
        kernel[1, 0] = -1
        kernel[1, 2] = -1
        kernel[0, 1] = -1
        kernel[2, 1] = -1
        kernel = Field(kernel.ravel(), kernel_grid)

        return generate_convolution_matrix(grid, kernel)
    else:
        raise NotImplementedError()

def make_derivative_matrix(grid, axis='x'):
    '''Make the derivative operator using the central difference approximation.

    Parameters
    ----------
    grid : Grid
        The grid for which the derivative matrix is calculated.
    axis : string
        The axis for which the convolution kernel is calculated default is 'x'.

    Returns
    -------
    array_like
        The convolution matrix.
    '''
    if grid.is_('cartesian') and grid.is_separated and grid.is_regular:
        num_x = 3
        num_y = 3
        kernel_grid = make_uniform_grid((num_x, num_y), (num_x * grid.delta[0], num_y * grid.delta[1]), has_center=True)
        kernel = kernel_grid.zeros()
        kernel = kernel.shaped

        if axis == 'x':
            kernel[1, 0] = -1 / (2 * grid.delta[1])
            kernel[1, 2] = 1 / (2 * grid.delta[1])
        elif axis == 'y':
            kernel[0, 1] = -1 / (2 * grid.delta[0])
            kernel[2, 1] = 1 / (2 * grid.delta[0])
        else:
            raise NotImplementedError()

        kernel = Field(kernel.ravel(), kernel_grid)
        return generate_convolution_matrix(grid, kernel)
    else:
        raise NotImplementedError()



from ..optics import OpticalSystem, NoiselessDetector, EmptyOpticalElement, FrameCorrector

class WavefrontSensorOptics(OpticalSystem):
    '''The optics for a wavefront sensor.

    This object serves as the base class for all optics in a wavefront sensor.
    '''
    pass

class WavefrontSensorEstimator(object):
    '''The estimator for a wavefront sensor.

    This object serves as the base class for all estimators in a wavefront sensor.
    It calculates from a list of images what the wavefront is. This wavefront can be
    anything that responds to wavefront changes, such as slopes, pupil amplitude, etc...
    '''
    def estimate(self, images):
        '''Estimate the wavefront from `images`.

        Parameters
        ----------
        images : list
            The list of images on which to base the estimate.

        Returns
        -------
        Field or ndarray
            The estimate of the wavefront.
        '''
        raise NotImplementedError()

class WavefrontSensor(object):
    '''A complete wavefront sensor.

    This convenience class consolidates the optics, detector, frame correction
    and estimation parts of a wavefront sensor.

    .. caution::
        This class only works with estimators that require only one image.
        For estimators with more than one required image, the wavefront
        sensor must be handcoded.

    Parameters
    ----------
    wfs_optics : WavefrontSensorOptics object
        The optics of the wavefront sensor. If it is not supplied, no optics
        will be used.
    detector : Detector object
        The detector of the wavefront sensor. If it is not supplied, a NoiselessDetector
        will be used.
    frame_correction : FrameCorrector object
        The frame correction of the detector. If it is not supplied, no frame
        correction will be used.
    wfs_estimator : WavefrontEstimator
        The wavefront estimator. If it is not supplied, the corrected image will
        be returned instead.
    '''
    def __init__(self, wfs_optics=None, detector=None, frame_corrector=None, wfs_estimator=None):
        self.wfs_optics = wfs_optics
        self.detector = detector
        self.frame_corrector = frame_corrector
        self.wfs_estimator = wfs_estimator

    @property
    def detector(self):
        '''The used detector.
        '''
        return self._detector

    @detector.setter
    def detector(self, detector):
        if detector is None:
            self._detector = NoiselessDetector()
        else:
            self._detector = detector

    @property
    def wfs_optics(self):
        '''The used wavefront sensor optics.
        '''
        return self._wfs_optics

    @wfs_optics.setter
    def wfs_optics(self, wfs_optics):
        if wfs_optics is None:
            self._wfs_optics = EmptyOpticalElement()
        else:
            self._wfs_optics = wfs_optics

    @property
    def frame_corrector(self):
        '''The used frame corrector.
        '''
        return self._frame_corrector

    @frame_corrector.setter
    def frame_corrector(self, frame_corrector):
        if frame_corrector is None:
            self._frame_corrector = FrameCorrector()
        else:
            self._frame_corrector = frame_corrector

    def integrate(self, wavefront, dt, weight=1):
        '''Integrate the wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront will be propagated through the wavefront sensor optics
            and fall onto the detector.
        dt : scalar
            The integration time in units of time.
        weight : scalar
            Weight of every unit of integration time.
        '''
        wf = self.wfs_optics.forward(wavefront)
        self.detector.integrate(wf, dt, weight)

    def read_out(self):
        '''Estimate the wavefront from the read-out detector image.

        If there is no wavefront estimator avaiable, the corrected detector
        image will be returned instead.

        Returns
        -------
        Field or ndarray
            The estimate of the wavefront sensor. If the wavefront sensor estimator
            is not give, the corrected detector image will be given instead.
        '''
        img = self.detector.read_out()
        corrected_img = self.frame_corrector.correct(img)

        if self.wfs_estimator is None:
            return corrected_img

        wfs_response = self.wfs_estimator.estimate([corrected_img])
        return wfs_response

    def __call__(self, wavefront, dt=1, weight=1):
        '''Convenience function to integrate and read out the wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront will be propagated through the wavefront sensor optics
            and fall onto the detector.
        dt : scalar
            The integration time in units of time.
        weight : scalar
            Weight of every unit of integration time.

        Returns
        -------
        Field or ndarray
            The estimate of the wavefront sensor. If the wavefront sensor estimator
            is not give, the corrected detector image will be given instead.
        '''
        self.integrate(wavefront, dt, weight)
        return self.read_out()



from .field import imshow_field
from ..field import Field
from ..config import Configuration

import numpy as np

def imshow_psf(
        psf, grid=None, vmin=1e-8, vmax=1e-1, scale='log',
        cmap=None, title=None, normalization='none',
        crosshairs=False, mark_centroid=False, colorbar=True,
        colorbar_orientation='vertical', spatial_resolution=1,
        ax=None, ticks=None, **kwargs):
    '''Display a PSF in a nice, consistent format.

    Parameters
    ----------
    psf : Field or Wavefront
        The point spread function that we want to display.
    grid : Grid or None
        If we want to redefine the grid of the psf.
    vmin : scalar
        The minimum value on the colorbar.
    vmax : scalar
        The maximum value on the colorbar.
    scale : ['linear', 'log', 'logarithmic']
        The scale of the image.
    cmap : matplotlib.cm.Colormap instance
        The colormap to use. If default, it will be taken from the configuration file.
    title : string, optional
        Give the image a title. If this is not given, no title will be displayed.
    normalization : ['peak', 'total', 'none'] or scalar
        The type of normalization to apply. 'peak' normalizes by the maximum intensity.
        'total' normalizes by the total intensity, 'none' doesn't do any normalization.
        If this is a scalar, the PSF will be divided by this number before displaying it.
    crosshairs : boolean
        Whether to draw crosshairs at the image center. Default: False.
    mark_centroid : boolean
        Whether to draw a small crosshair at the centroid of the PSF. The centroid will be
        computed with TODO Default: False.
    colorbar : boolean
        Whether to draw a colorbar on the image. Default: True.
    colorbar_orientation : ['horizontal', 'vertical']
        How the colorbar should be oriented.
    spatial_resolution : scalar
        This will be the size of a unit cell on the axes.
    ax : matplotlib axes
        The axes which to draw on. If it is not given, the current axes will be used.
    ticks : array_like
        The ticks on the added colorbar.
    kwargs : dictionary
        Any keyword arguments will be sent to `imshow_field()`.

    Returns
    -------
    Image
        The created image.
    '''
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    if ax is None:
        ax = plt.gca()

    if hasattr(psf, 'wavelength'):
        psf = psf.power

    if grid is not None:
        psf = Field(psf, grid)

    if normalization.lower() == 'peak':
        psf_norm = psf.max()
    elif normalization.lower() == 'total':
        psf_norm = np.sum(psf * psf.grid.weights)
    elif normalization.lower() == 'none':
        psf_norm = 1
    else:
        psf_norm = normalization

    if cmap is None:
        cmap = Configuration().plotting.psf_colormap

    img = psf / psf_norm

    if scale.lower() == 'linear':
        norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    elif scale.lower() == 'log' or scale.lower() == 'logarithmic':
        norm = mpl.colors.LogNorm(vmin=vmin, vmax=vmax)
    else:
        raise ValueError('Scale needs to be one of ["linear", "log", "logarithmic"].')

    im = imshow_field(
        img, norm=norm, cmap=cmap,
        grid_units=spatial_resolution, ax=ax, **kwargs
    )

    if crosshairs:
        ax.axhline(0, ls=':', color='k')
        ax.axvline(0, ls=':', color='k')

    if title:
        ax.set_title(title)

    if colorbar:
        cb = plt.colorbar(im, ax=ax, orientation=colorbar_orientation)

        if scale.lower() in ['log', 'logarithmic']:
            if ticks is None:
                ticks = np.logspace(np.log10(vmin), np.log10(vmax), int(np.round(np.log10(vmax / vmin) + 1)))

                if colorbar_orientation == 'horizontal' and vmax == 1e-1 and vmin == 1e-8:
                    # Use better looking ticks
                    ticks = [1e-8, 1e-6, 1e-4, 1e-2, 1e-1]

            cb.set_ticks(ticks)
            cb.set_ticklabels(ticks)

        if normalization.lower() == 'peak':
            cb.set_label('Intensity relative to peak')
        elif normalization.lower() == 'total':
            cb.set_label('Fractional intensity per pixel')

    if mark_centroid:
        raise NotImplementedError()

    return im

def imshow_pupil_phase(
        pupil_phase, grid=None, phase_limits=None, vmin=None, vmax=None, cmap=None,
        colorbar=True, colorbar_orientation='vertical', title=None, crosshairs=False,
        remove_piston=False, ax=None, **kwargs):
    '''Display a pupil phase pattern in a nice, consistent format.

    Parameters
    ----------
    pupil_phase : Field or Wavefront
        The pupil that we want to display. If it is a wavefront or complex, the phase will be used
        as the phase, and the amplitude will be applied as mask.
    grid : Grid or None
        If we want to redefine the grid of the pupil.
    phase_limits : scalar
        The minimum and maximum of the pupil phase. If it is not given, and vmin and vmax are
        not given either, it will be estimated from the data. If the range is sufficiently close
        to 2 pi, the range will be extended to 2 pi. The vmin and vmax, unless overridden, will
        be set to -phase_limits and phase_limits, so the phase will be centered around zero.
    vmin : scalar, optional
        The minimum of the colorbar.
    vmax : scalar, optional
        The maximum of the colorbar.
    cmap : matplotlib.cm.Colormap instance
        The colormap to use. If default, it will be taken from the configuration file.
    colorbar : boolean
        Whether to draw a colorbar on the image. Default: True.
    colorbar_orientation : ['horizontal', 'vertical']
        How the colorbar should be oriented.
    title : string, optional
        Give the image a title. If this is not given, no title will be displayed.
    crosshairs : boolean
        Whether to draw crosshairs at the image center. Default: False.
    remove_piston : boolean
        Whether to remove piston from the phase. Default: False.
    ax : matplotlib axes
        The axes which to draw on. If it is not given, the current axes will be used.
    kwargs : dictionary
        Any keyword arguments will be sent to `imshow_field()`.
    '''
    import matplotlib.pyplot as plt

    if ax is None:
        ax = plt.gca()

    if hasattr(pupil_phase, 'wavelength'):
        pupil_phase = pupil_phase.electric_field

    if grid is not None:
        pupil_phase = Field(pupil_phase, grid)

    if np.iscomplexobj(pupil_phase):
        phase = Field(np.angle(pupil_phase), pupil_phase.grid)
        mask = np.abs(pupil_phase)
    else:
        phase = pupil_phase
        mask = None

    if remove_piston:
        complex_field = np.exp(1j * phase)
        if mask is not None:
            complex_field *= mask

        complex_field /= np.mean(complex_field)
        phase = Field(np.angle(complex_field), phase.grid)

    if phase_limits is None:
        phase_limits = max(-phase[mask > 0].min(), phase[mask > 0].max())

        if np.abs(phase_limits - np.pi) < 0.5:
            phase_limits = np.pi

        if phase_limits < 1e-10:
            phase_limits = 1e-10

    if vmin is None and vmax is None:
        vmin = -phase_limits
        vmax = phase_limits

    if cmap is None:
        cmap = Configuration().plotting.pupil_phase_colormap

    im = imshow_field(phase, ax=ax, cmap=cmap, mask=mask, vmin=vmin, vmax=vmax, **kwargs)

    if crosshairs:
        ax.axhline(0, ls=':', color='k')
        ax.axvline(0, ls=':', color='k')

    if title:
        ax.set_title(title)

    if colorbar:
        cb = plt.colorbar(im, ax=ax, orientation=colorbar_orientation)
        cb.set_label('Phase [rad]')

    return im



import os
import shutil
import base64
from subprocess import Popen, PIPE
import io

import matplotlib as mpl
import matplotlib.pyplot as plt
import imageio.v2 as imageio

from ..config import Configuration

def _data_to_img(data, cmap, copy=False):
    if cmap is None:
        if copy:
            return data.copy()

        return data

    try:
        cmap = mpl.colormaps.get_cmap(cmap)
    except AttributeError:
        # For Matplotlib <3.5.
        cmap = mpl.cm.get_cmap(cmap)

    return cmap(data, bytes=True)

class FrameWriter(object):
    '''A writer of frames from Matplotlib figures.

    This class writes out individual frames to the specified directory. No
    animation will be created for these frames.

    Parameters
    ----------
    path : string
        The directory to which to write all frames. This path will be
        created if it does not exist.
    framerate : integer
        Ignored, but provided for cohesion with other animation writers.
    filename : string
        The filename for each of the frames. This will be formatted using
        filename.format(frame_number). Default: '{:05d}.png'.
    '''
    def __init__(self, path, framerate=15, filename='{:05d}.png'):
        self.path = path
        self.filename = filename

        self.num_frames = 0

        if not os.path.exists(self.path):
            os.mkdir(self.path)

        self.is_closed = False

    def add_frame(self, fig=None, data=None, cmap=None, dpi=None):
        '''Add a frame to the animation.

        Parameters
        ----------
        fig : Matplotlib figure
            The Matplotlib figure acting as the animation frame.
        data : ndarray
            The image data array acting as the animation frame.
        cmap : Matplotlib colormap
            The optional colormap for the image data.
        dpi : integer or None
            The number of dots per inch with which to save the matplotlib figure.
            If it is not given, the default Matplotlib dpi will be used.

        Raises
        ------
        RuntimeError
            If the function was called on a closed FrameWriter.
        '''
        if self.is_closed:
            raise RuntimeError('Attempted to add a frame to a closed GifWriter.')

        dest = os.path.join(self.path, self.filename.format(self.num_frames))

        if data is None:
            if fig is None:
                fig = plt.gcf()

            facecolor = list(fig.get_facecolor())
            facecolor[3] = 1

            fig.savefig(dest, transparent=False, dpi=dpi, facecolor=facecolor)
        else:
            img = _data_to_img(data, cmap)

            imageio.imwrite(dest, img)

        self.num_frames += 1

    def close(self):
        '''Closes the animation writer.

        Makes sure that no more frames can be written to the directory.
        '''
        self.is_closed = True

class GifWriter(object):
    '''A writer of gif files from Matplotlib figures.

    .. warning::
        This class used to write out individual frames to a directory,
        before converting this into a gif file. This is now done internally.

    Parameters
    ----------
    filename : string
        The path and filename of the gif.
    framerate : integer
        The number of frames per second of the generated gif file.
    '''
    def __init__(self, filename, framerate=15):
        self.is_closed = False
        self.filename = filename
        self.framerate = framerate

        self._frames = []

    @property
    def num_frames(self):
        return len(self._frames)

    def __del__(self):
        try:
            self.close()
        except Exception:
            import warnings
            warnings.warn('Something went wrong while closing the GifWriter...', RuntimeWarning)

    def add_frame(self, fig=None, data=None, cmap=None, dpi=None):
        '''Add a frame to the animation.

        Parameters
        ----------
        fig : Matplotlib figure
            The Matplotlib figure acting as the animation frame.
        data : ndarray
            The image data array acting as the animation frame.
        cmap : Matplotlib colormap
            The optional colormap for the image data.
        dpi : integer or None
            The number of dots per inch with which to save the matplotlib figure.
            If it is not given, the default Matplotlib dpi will be used.

        Raises
        ------
        RuntimeError
            If the function was called on a closed GifWriter.
        '''
        if self.is_closed:
            raise RuntimeError('Attempted to add a frame to a closed GifWriter.')

        if data is None:
            if fig is None:
                fig = plt.gcf()

            facecolor = list(fig.get_facecolor())
            facecolor[3] = 1

            buf = io.BytesIO()

            fig.savefig(buf, transparent=False, dpi=dpi, facecolor=facecolor)

            frame = imageio.imread(buf.getvalue())
        else:
            frame = _data_to_img(data, cmap, copy=True)

        self._frames.append(frame)

    def convert(self):
        '''Convert all frames into a gif file.
        '''
        imageio.mimsave(self.filename, self._frames, duration=1 / self.framerate)

    def close(self):
        '''Close the animation and create the final gif file.
        '''
        try:
            if not self.is_closed:
                self.convert()
        finally:
            self.is_closed = True

class FFMpegWriter(object):
    '''A writer of video files from Matplotlib figures.

    This class uses FFMpeg as the basis for writing frames to a video file.

    Parameters
    ----------
    filename : string
        The filename of the generated video file.
    codec : string
        The codec for FFMpeg to use. If it is not given, a suitable codec
        will be guessed based on the file extension.
    framerate : integer
        The number of frames per second of the generated video file.
    quality : string
        The quality of the encoding for lossy codecs. Please refer to FFMpeg documentation.
    preset : string
        The preset for the quality of the encoding. Please refer to FFMpeg documentation.

    Raises
    ------
    ValueError
        If the codec was not given and could not be guessed based on the file extension.
    RuntimeError
        If something went wrong during initialization of the call to FFMpeg. Most likely,
        FFMpeg is not installed and/or not available from the commandline.
    '''
    def __init__(self, filename, codec=None, framerate=24, quality=None, preset=None):
        if codec is None:
            extension = os.path.splitext(filename)[1]
            if extension == '.mp4' or extension == '.avi':
                codec = 'libx264'
            elif extension == '.webm':
                codec = 'libvpx-vp9'
            else:
                raise ValueError('No codec was given and it could not be guessed based on the file extension.')

        self.is_closed = True
        self.filename = filename
        self.codec = codec
        self.framerate = framerate

        ffmpeg_path = Configuration().plotting.ffmpeg_path
        if ffmpeg_path is None:
            ffmpeg_path = 'ffmpeg'

        if shutil.which(ffmpeg_path) is None:
            raise RuntimeError('ffmpeg was not found. Did you install it and is it accessible, either from PATH or from the HCIPy configuration file?')

        if codec == 'libx264':
            if quality is None:
                quality = 10

            if preset is None:
                preset = 'veryslow'

            command = [
                ffmpeg_path, '-y', '-nostats', '-v', 'quiet', '-f', 'image2pipe',
                '-vcodec', 'png', '-r', str(framerate), '-threads', '0', '-i', '-',
                '-vcodec', 'libx264', '-pix_fmt', 'yuv420p', '-preset', preset, '-r',
                str(framerate), '-crf', str(quality), filename
            ]
        elif codec == 'mpeg4':
            if quality is None:
                quality = 4

            command = [
                ffmpeg_path, '-y', '-nostats', '-v', 'quiet', '-f', 'image2pipe',
                '-vcodec', 'png', '-r', str(framerate), '-threads', '0', '-i', '-',
                '-vcodec', 'mpeg4', '-q:v', str(quality), '-r', str(framerate), filename
            ]
        elif codec == 'libvpx-vp9':
            if quality is None:
                quality = 30

            command = [
                ffmpeg_path, '-y', '-nostats', '-v', 'quiet', '-f', 'image2pipe',
                '-vcodec', 'png', '-r', str(framerate), '-threads', '0', '-i', '-'
            ]

            if quality < 0:
                command.extend(['-vcodec', 'libvpx-vp9', '-lossless', '1', '-r', str(framerate), filename])
            else:
                command.extend(['-vcodec', 'libvpx-vp9', '-crf', str(quality), '-b:v', '0', '-r', str(framerate), filename])
        else:
            raise ValueError('Codec unknown.')

        try:
            self.p = Popen(command, stdin=PIPE)
        except OSError:
            raise RuntimeError('Something went wrong when opening FFMpeg.')

        self.is_closed = False

    def __del__(self):
        try:
            self.close()
        except Exception:
            import warnings
            warnings.warn('Something went wrong while closing FFMpeg...', RuntimeWarning)

    def add_frame(self, fig=None, data=None, cmap=None, dpi=None):
        '''Add a frame to the animation.

        Parameters
        ----------
        fig : Matplotlib figure
            The Matplotlib figure acting as the animation frame.
        data : ndarray
            The image data array acting as the animation frame.
        cmap : Matplotlib colormap
            The optional colormap for the image data.
        dpi : integer or None
            The number of dots per inch with which to save the matplotlib figure.
            If it is not given, the default Matplotlib dpi will be used.

        Raises
        ------
        RuntimeError
            If the function was called on a closed FFMpegWriter.
        '''
        if self.is_closed:
            raise RuntimeError('Attempted to add a frame to a closed FFMpegWriter.')

        if data is None:
            if fig is None:
                fig = plt.gcf()

            facecolor = list(fig.get_facecolor())
            facecolor[3] = 1

            fig.savefig(self.p.stdin, format='png', transparent=False, dpi=dpi, facecolor=facecolor)
        else:
            img = _data_to_img(data, cmap)

            imageio.imwrite(self.p.stdin, img, format='png')

    def close(self):
        '''Close the animation writer and finish the video file.

        This closes the FFMpeg call.
        '''
        if not self.is_closed:
            self.p.stdin.close()
            self.p.wait()
            self.p = None
        self.is_closed = True

    def _repr_html_(self):
        '''Get an HTML representation of the generated video.

        Helper function for Jupyter notebooks. The video will be inline embedded in an
        HTML5 video tag using base64 encoding. This is not very efficient, so only use this
        for small video files.

        The FFMpegWriter must be closed for this function to work.

        Raises
        ------
        RuntimeError
            If the call was made on an open FFMpegWriter.
        '''
        if not self.is_closed:
            raise RuntimeError('Attempted to show the generated movie on an opened FFMpegWriter.')

        video = open(self.filename, 'rb').read()
        video = base64.b64encode(video).decode('ascii').rstrip()

        if self.filename.endswith('.mp4'):
            mimetype = 'video/mp4'
        elif self.filename.endswith('.webm'):
            mimetype = 'video/webm'
        elif self.filename.endswith('.avi'):
            mimetype = 'video/avi'
        else:
            raise RuntimeError('Mimetype could not be guessed.')

        output = '''<video controls><source src="data:{0};base64,{1}" type="{0}">Your browser does not support the video tag.</video>'''
        output = output.format(mimetype, video)

        return output



import numpy as np
from .optical_element import OpticalElement, AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward

class Apodizer(AgnosticOpticalElement):
    '''A thin apodizer.

    This apodizer can apodize both in phase and amplitude.

    Parameters
    ----------
    apodization : Field or scalar or function of wavelength
        The apodization that we want to apply to any input wavefront.
    '''
    def __init__(self, apodization):
        self._apodization = apodization

        AgnosticOpticalElement.__init__(self, True, True)

    def make_instance(self, instance_data, input_grid, output_grid, wavelength):
        instance_data.apodization = self.evaluate_parameter(self.apodization, input_grid, output_grid, wavelength)

    @property
    def apodization(self):
        return self._apodization

    @apodization.setter
    def apodization(self, apodization):
        self._apodization = apodization

        self.clear_cache()

    def get_input_grid(self, output_grid, wavelength):
        return output_grid

    def get_output_grid(self, input_grid, wavelength):
        return input_grid

    @make_agnostic_forward
    def forward(self, instance_data, wavefront):
        wf = wavefront.copy()
        wf.electric_field *= instance_data.apodization

        return wf

    @make_agnostic_backward
    def backward(self, instance_data, wavefront):
        wf = wavefront.copy()
        wf.electric_field *= instance_data.apodization.conj()

        return wf

class PhaseApodizer(Apodizer):
    '''A phase-only thin apodizer.

    Parameters
    ----------
    phase : Field or scalar or function
        The phase apodization.
    '''
    def __init__(self, phase):
        self._phase = phase

        Apodizer.__init__(self, self.apodization)

    @property
    def apodization(self):
        return self.construct_function(lambda p: np.exp(1j * p), self.phase)

    @property
    def phase(self):
        return self._phase

    @phase.setter
    def phase(self, phase):
        self._phase = phase

        self.clear_cache()

class SurfaceApodizer(Apodizer):
    '''A transmissive sagged surface optic.

    The surface is simulated as a thin plate. Propagation effects due to the
    thickness of the plate are not included. The supplied refractive index
    may change as function of wavelength.

    Parameters
    ----------
    surface_sag : Field or scalar or function
        The sag in the surface.
    refractive_index : scalar or function
        The refractive index of the material of the plate.
    '''
    def __init__(self, surface_sag, refractive_index):
        self._surface_sag = surface_sag
        self._refractive_index = refractive_index

        Apodizer.__init__(self, self.apodization)

    @property
    def opd(self):
        return self.construct_function(lambda n, surf: (n - 1) * surf, self.refractive_index, self.surface_sag)

    optical_path_difference = opd

    @property
    def phase(self):
        return self.construct_function(lambda opd, wavelength: opd * 2 * np.pi / wavelength, self.opd)

    @property
    def apodization(self):
        return self.construct_function(lambda p: np.exp(1j * p), self.phase)

    @property
    def refractive_index(self):
        return self._refractive_index

    @refractive_index.setter
    def refractive_index(self, refractive_index):
        self._refractive_index = refractive_index

        self.clear_cache()

    @property
    def surface_sag(self):
        return self._surface_sag

    @surface_sag.setter
    def surface_sag(self, surface_sag):
        self._surface_sag = surface_sag

        self.clear_cache()

class ComplexSurfaceApodizer(OpticalElement):
    def __init__(self, amplitude, surface, refractive_index):
        self.amplitude = amplitude
        self.surface = surface
        self.refractive_index = refractive_index

    def phase_for(self, wavelength):
        '''Get the phase screen in radians at a certain wavelength.

        Parameters
        ----------
        wavelength : scalar
            The wavelength at which to calculate the phase screen.
        '''
        wavenumber = 2 * np.pi / wavelength

        opd = (self.refractive_index - 1) * self.surface

        return opd * wavenumber

    def forward(self, wavefront):
        opd = (self.refractive_index(wavefront.wavelength) - 1) * self.surface

        wf = wavefront.copy()
        wf.electric_field *= self.amplitude * np.exp(1j * opd * wf.wavenumber)

        return wf

    def backward(self, wavefront):
        opd = (self.refractive_index(wavefront.wavelength) - 1) * self.surface

        wf = wavefront.copy()
        wf.electric_field *= self.amplitude * np.exp(-1j * opd * wf.wavenumber)

        return wf

class MultiplexedComplexSurfaceApodizer(OpticalElement):
    def __init__(self, amplitude, surface, refractive_index):
        self.amplitude = amplitude
        self.surface = surface
        self.refractive_index = refractive_index

    def forward(self, wavefront):
        apodizer_mask = 0
        for amplitude, surface in zip(self.amplitude, self.surface):
            opd = (self.refractive_index(wavefront.wavelength) - 1) * surface
            apodizer_mask += amplitude * np.exp(1j * opd * wavefront.wavenumber)

        wf = wavefront.copy()
        wf.electric_field *= apodizer_mask
        return wf

    def backward(self, wavefront):
        apodizer_mask = 0
        for amplitude, surface in zip(self.amplitude, self.surface):
            opd = (self.refractive_index(wavefront.wavelength) - 1) * surface
            apodizer_mask += amplitude * np.exp(1j * opd * wavefront.wavenumber)

        wf = wavefront.copy()
        wf.electric_field /= apodizer_mask
        return wf



from ..optics import Apodizer, OpticalElement, Wavefront
from ..field import Field
from ..fourier import FastFourierTransform, make_fft_grid
from .._math import fft as _fft_module

import numpy as np

class KnifeEdgeLyotCoronagraph(OpticalElement):
    '''A Lyot-style coronagraph with a centered, knife-edge focal-plane mask.

    As the knife-edge is invariant along one axis, the Fourier transforms can
    be sped up by only taking them along one axis. Also, the zero-padding only
    needs to be done along that axis, reducing memory usage while doing the
    transform.

    The knife-edge is set along the y-axis, so perpendicular to the x-axis.

    Parameters
    ----------
    input_grid : Grid
        The grid of the wavefront that is to be propagated.
    q : scalar
        The amount of oversampling used for the knife edge.
    direction : string
        The direction of the transmissive part of the knife edge. This
        has to be one of {'+x', '-x', '+y', '-y'}.
    apodizer : OpticalElement or Field or None
        The pre-apodizer in the pupil before the focal-plane mask.
        If this is a Field, it will be converted into an apodizer.
        If this is None, no apodizer will be used.
    lyot_mask : OpticalElement or Field
        The Lyot stop in the pupil after the focal-plane mask. If
        this is a Field, it will be converted to an apodizer.
        If this is None, no Lyot stop will be used.
    '''
    def __init__(self, input_grid, q=8, direction='+x', apodizer=None, lyot_stop=None):
        self.direction = direction

        fft = FastFourierTransform(input_grid, q, 1)

        if 'x' in self.direction:
            self.cutout_input = (Ellipsis, fft.cutout_input[1])
            self.internal_shape = input_grid.shape[0], fft.internal_shape[1]
        else:
            self.cutout_input = (fft.cutout_input[0], Ellipsis)
            self.internal_shape = fft.internal_shape[0], input_grid.shape[1]

        # Create the knife-edge focal-plane mask along the x-axis.
        focal_mask_grid = make_fft_grid(input_grid, q).scaled(1.0 / (2 * np.pi))

        x = focal_mask_grid.separated_coords[0 if 'x' in self.direction else 1]
        self.focal_mask = 0.5 + 0.5 * np.sign(x) * (-1 if '-' in self.direction else 1)

        # Pre-shift the mask to speed up propagations
        # FIXME: not necessary when 1D FFTs on 2D grids are implemented.
        self.focal_mask = np.fft.fftshift(self.focal_mask)

        if apodizer is not None and not hasattr(apodizer, 'input_grid'):
            apodizer = Apodizer(apodizer)
        self.apodizer = apodizer

        if lyot_stop is not None and not hasattr(lyot_stop, 'input_grid'):
            lyot_stop = Apodizer(lyot_stop)
        self.lyot_stop = lyot_stop

    def forward(self, wavefront):
        '''Propagate a wavefront forward through the knife-edge coronagraph.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        if self.apodizer is not None:
            wavefront = self.apodizer.forward(wavefront)

        ap = np.zeros(self.internal_shape, dtype='complex')
        ap[tuple(self.cutout_input)] = wavefront.electric_field.shaped

        # FIXME: use 1D FFTs on 2D grids implementation when available, instead of this.
        if 'x' in self.direction:
            axis = 1
            focal_mask = self.focal_mask[np.newaxis, :]
        else:
            axis = 0
            focal_mask = self.focal_mask[:, np.newaxis]

        post_coro = _fft_module.ifft(_fft_module.fft(ap, axis=axis) * focal_mask, axis=axis)
        post_coro = Field(post_coro[self.cutout_input].ravel(), wavefront.electric_field.grid)

        wavefront = Wavefront(post_coro, wavefront.wavelength)

        if self.lyot_stop is not None:
            wavefront = self.lyot_stop.forward(wavefront)

        return wavefront

    def backward(self, wavefront):
        '''Propagate a wavefront backward through the knife-edge coronagraph.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        if self.lyot_stop is not None:
            wavefront = self.lyot_stop.backward(wavefront)

        ap = np.zeros(self.internal_shape, dtype='complex')
        ap[tuple(self.cutout_input)] = wavefront.electric_field.shaped

        # FIXME: use 1D FFTs on 2D grids implementation when available, instead of this.
        if 'x' in self.direction:
            axis = 1
            focal_mask = self.focal_mask[np.newaxis, :]
        else:
            axis = 0
            focal_mask = self.focal_mask[:, np.newaxis]

        post_coro = _fft_module.ifft(_fft_module.fft(ap, axis=axis) * focal_mask, axis=axis)
        post_coro = Field(post_coro[self.cutout_input].ravel(), wavefront.electric_field.grid)

        wavefront = Wavefront(post_coro, wavefront.wavelength)

        if self.apodizer is not None:
            wavefront = self.apodizer.backward(wavefront)

        return wavefront



import datetime
import io
import pickle

import numpy as np
import asdf
from astropy import wcs
from astropy.io import fits

from ..field import Grid, Field
from ..mode_basis import ModeBasis

try:
    from .._version import version as _version
except ImportError:
    _version = ''


def _asdf_to_bintable(af, *args, **kwargs):
    buffer = io.BytesIO()
    af.write_to(buffer, *args, **kwargs)
    buffer.seek(0)

    data = np.array(buffer.getbuffer(), dtype=np.uint8)[None, :]
    fmt = f"{len(data[0])}B"
    column = fits.Column(array=data, format=fmt, name="ASDF_METADATA")
    return fits.BinTableHDU.from_columns([column], name="ASDF")


def _bintable_to_asdf(bintable, *args, **kwargs):
    return asdf.open(io.BytesIO(bintable.data), *args, **kwargs)


def read_fits(filename, extension=0):
    '''Read an array from a fits file.

    Parameters
    ----------
    filename : string
        The filename of the file to read. This can include a path.
    extension : integer
        The extension of the fits file that is being read.

    Returns
    -------
    ndarray
        The ndarray read from the fits file.
    '''
    return fits.getdata(filename, extension).copy()

def write_fits(data, filename, shape=None, overwrite=True):
    '''Write the data to a fits-file.

    If the data is a Field with SeparatedCoords, the shaped field will be written to the
    fits-file, to allow for easier viewing with external tools.

    Parameters
    ----------
    data : ndarray or Field
        The ndarray or Field to write to the fits file.
    filename : string
        The filename of the newly created file. This can include a path.
    shape : ndarray or None
        The shape to which to reshape the data. If this is given, it will override a potential
        shape from the grid accompaning the field.
    overwrite : boolean
        Whether to overwrite the fits-file if it already exists.
    '''
    hdu = None

    if shape is not None:
        hdu = fits.PrimaryHDU(data.reshape(shape))
    elif hasattr(data, 'grid'):
        if data.grid.is_separated:
            hdu = fits.PrimaryHDU(data.shaped)

    if hdu is None:
        hdu = fits.PrimaryHDU(data)

    hdu.writeto(filename, overwrite=True)

def _guess_file_format(filename):
    '''Guess the file format from the filename.

    Parameters
    ----------
    filename : string
        The filename of the file.

    Returns
    -------
    string
        The file type. This is one of ["asdf", "fits"].
    '''
    if filename.endswith('asdf'):
        return 'asdf'
    elif filename.endswith('fits') or filename.endswith('fits.gz'):
        return 'fits'
    elif filename.endswith('pkl') or filename.endswith('pickle'):
        return 'pickle'
    else:
        return None

def _make_metadata(file_type):
    '''Make a metadata section for file output.

    This contains the version of HCIPy used to write the file, the current time and
    date and the file type.

    Parameters
    ----------
    file_type : string
        The identifier string for what is going to be stored in the file.

    Returns
    -------
    dictionary
        The metadata.
    '''
    tree = {
        'meta': {
            'author': 'HCIPy %s' % _version,
            'date_utc': datetime.datetime.now(datetime.timezone.utc).isoformat(),
            'file_type': file_type
        }
    }

    return tree

def read_grid(filename, fmt=None):
    '''Read a grid from a file.

    Parameters
    ----------
    filename : string
        The path of the file you want to read the grid from.
    fmt : string
        The file format. If it is not given, the file format will be guessed from the file extension.

    Returns
    -------
    Grid
        The read grid.

    Raises
    ------
    ValueError
        If the file format could not be guessed from the file extension.
    NotImplementedError
        If the file format was not yet implemented.
    '''
    if fmt is None:
        fmt = _guess_file_format(filename)

        if fmt is None:
            raise ValueError('Format not given and could not be guessed based on the file extension.')

    if fmt == 'fits':
        with fits.open(filename, memmap=False) as f:
            af = _bintable_to_asdf(f['ASDF'])
            return Grid.from_dict(af.tree['grid'])
    elif fmt == 'asdf':
        with asdf.open(filename, copy_arrays=True) as f:
            return Grid.from_dict(f.tree['grid'])
    elif fmt == 'pickle':
        with open(filename, 'rb') as f:
            return pickle.load(f)
    else:
        raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)

def write_grid(grid, filename, fmt=None, overwrite=True):
    '''Write a grid to a file.

    Parameters
    ----------
    grid : Grid
        The grid to write to the file.
    filename : string
        The path of the file you want to write to.
    fmt : string
        The file format. If it is not given, the file format will be guessed based on the file extension.
    overwrite : boolean
        Whether to overwrite the file, if it already exists.

    Raises
    ------
    NotImplementedError
        If the file format was not yet implemented.
    '''
    if fmt is None:
        fmt = _guess_file_format(filename)

        if fmt is None:
            raise ValueError('Format not given and could not be guessed based on the file extension.')

    tree = _make_metadata('grid')
    tree['grid'] = grid.to_dict()

    if fmt == 'asdf':
        target = asdf.AsdfFile(tree)
        target.write_to(filename, all_array_compression='zlib')
    elif fmt == 'fits':
        hdulist = fits.HDUList()
        hdulist.append(_asdf_to_bintable(asdf.AsdfFile(tree), all_array_compression='zlib'))
        hdulist.writeto(filename, overwrite=overwrite)
    elif fmt == 'pickle':
        with open(filename, 'wb') as f:
            pickle.dump(grid, f)
    else:
        raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)

def read_field(filename, fmt=None):
    '''Read a field from a file.

    Parameters
    ----------
    filename : string
        The path of the file you want to read the field from.
    fmt : string
        The file format. If it is not given, the file format will be guessed from the file extension.

    Returns
    -------
    Field
        The read field.

    Raises
    ------
    ValueError
        If the file format could not be guessed from the file extension.
    NotImplementedError
        If the file format was not yet implemented.
    '''
    if fmt is None:
        fmt = _guess_file_format(filename)

        if fmt is None:
            raise ValueError('Format not given and could not be guessed based on the file extension.')

    if fmt == 'asdf':
        with asdf.open(filename, copy_arrays=True) as f:
            return Field.from_dict(f.tree['field'])
    elif fmt == 'fits':
        with fits.open(filename, memmap=False) as f:
            af = _bintable_to_asdf(f['ASDF'])
            tree = af.tree['field']
            if f[0].data is not None:
                tree['values'] = f[0].data

            if 'grid' in tree:
                grid = Grid.from_dict(tree['grid'])
                new_shape = np.concatenate((tree['values'].shape[:-grid.ndim], [grid.size])).astype('int')
                tree['values'] = tree['values'].reshape(new_shape)

            return Field.from_dict(tree).reshape(new_shape)
    elif fmt == 'pickle':
        with open(filename, 'rb') as f:
            return pickle.load(f)
    else:
        raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)

def write_field(field, filename, fmt=None, overwrite=True):
    '''Write a field to a file.

    Parameters
    ----------
    field : Field
        The field to write to the file.
    filename : string
        The path of the file you want to write to.
    fmt : string
        The file format. If it is not given, the file format will be guessed based on the file extension.
    overwrite : boolean
        Whether to overwrite the file, if it already exists.

    Raises
    ------
    NotImplementedError
        If the file format was not yet implemented.
    '''
    if fmt is None:
        fmt = _guess_file_format(filename)

        if fmt is None:
            raise ValueError('Format not given and could not be guessed based on the file extension.')

    tree = _make_metadata('field')
    tree['field'] = field.to_dict()

    if fmt == 'asdf':
        target = asdf.AsdfFile(tree)
        target.write_to(filename, all_array_compression='zlib')
    elif fmt == 'fits':
        hdulist = fits.HDUList()

        if field.grid.is_separated:
            hdulist.append(fits.ImageHDU(np.ascontiguousarray(field.shaped)))
            del tree['field']['values']

        if field.grid.is_regular:
            w = wcs.WCS(naxis=field.grid.ndim)
            w.wcs.crpix = np.ones(field.grid.ndim)
            w.wcs.cdelt = field.grid.delta
            w.wcs.crval = field.grid.zero
            w.wcs.ctype = ['X', 'Y', 'Z', 'W'][:field.grid.ndim]

            hdulist[0].header.update(w.to_header())

        hdulist.append(_asdf_to_bintable(asdf.AsdfFile(tree), all_array_compression='zlib'))
        hdulist.writeto(filename, overwrite=overwrite)
    elif fmt == 'pickle':
        with open(filename, 'wb') as f:
            pickle.dump(field, f)
    else:
        raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)

def read_mode_basis(filename, fmt=None):
    '''Read a mode basis from a file.

    Parameters
    ----------
    filename : string
        The path of the file you want to read the mode basis from.
    fmt : string
        The file format. If it is not given, the file format will be guessed from the file extension.

    Returns
    -------
    ModeBasis
        The read mode basis.

    Raises
    ------
    ValueError
        If the file format could not be guessed from the file extension.
    NotImplementedError
        If the file format was not yet implemented.
    '''
    if fmt is None:
        fmt = _guess_file_format(filename)

        if fmt is None:
            raise ValueError('Format not given and could not be guessed based on the file extension.')

    if fmt == 'asdf':
        with asdf.open(filename, copy_arrays=True) as f:
            return ModeBasis.from_dict(f.tree['mode_basis'])
    elif fmt == 'fits':
        with fits.open(filename, memmap=False) as f:
            af = _bintable_to_asdf(f['ASDF'])
            tree = af.tree['mode_basis']

            if f[0].data is not None:
                grid = Grid.from_dict(tree['grid'])
                modes = f[0].data

                old_shape = np.concatenate((modes.shape[:-grid.ndim], [grid.size]))
                tree['transformation_matrix'] = modes.reshape(old_shape).T

            return ModeBasis.from_dict(tree)
    elif fmt == 'pickle':
        with open(filename, 'rb') as f:
            return pickle.load(f)
    else:
        raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)

def write_mode_basis(mode_basis, filename, fmt=None, overwrite=True):
    '''Write a mode basis to a file.

    Parameters
    ----------
    mode_basis : ModeBasis
        The mode basis to write to the file.
    filename : string
        The path of the file you want to write to.
    fmt : string
        The file format. If it is not given, the file format will be guessed based on the file extension.
    overwrite : boolean
        Whether to overwrite the file, if it already exists.

    Raises
    ------
    NotImplementedError
        If the file format was not yet implemented.
    '''
    if fmt is None:
        fmt = _guess_file_format(filename)

        if fmt is None:
            raise ValueError('Format not given and could not be guessed based on the file extension.')

    tree = _make_metadata('mode_basis')
    tree['mode_basis'] = mode_basis.to_dict()

    if fmt == 'asdf':
        target = asdf.AsdfFile(tree)
        target.write_to(filename, all_array_compression='zlib')
    elif fmt == 'fits':
        hdulist = fits.HDUList()

        if mode_basis.grid and mode_basis.grid.is_separated:
            new_shape = np.concatenate(([-1], mode_basis.grid.shape))
            modes = np.ascontiguousarray(mode_basis.to_dense().transformation_matrix.T.reshape(new_shape))

            hdulist.append(fits.ImageHDU(np.asarray(modes)))

            w = wcs.WCS(naxis=mode_basis.grid.ndim)
            w.wcs.crpix = np.ones(mode_basis.grid.ndim)
            w.wcs.cdelt = mode_basis.grid.delta
            w.wcs.crval = mode_basis.grid.zero
            w.wcs.ctype = ['X', 'Y', 'Z', 'W'][:mode_basis.grid.ndim]

            hdulist[0].header.update(w.to_header())

            del tree['mode_basis']['transformation_matrix']

        hdulist.append(_asdf_to_bintable(asdf.AsdfFile(tree), all_array_compression='zlib'))
        hdulist.writeto(filename, overwrite=overwrite)
    elif fmt == 'pickle':
        with open(filename, 'wb') as f:
            pickle.dump(mode_basis, f)
    else:
        raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)



import numpy as np
import scipy.sparse
import scipy.sparse.linalg

from ..mode_basis import ModeBasis

class SVD(object):
    '''The Singular Value Decomposition for the provided matrix.

    This class wraps two versions of the SVD in numpy and scipy, and provides
    easy access to singular modes (as mode bases) and allows for calculation
    of the SVD for a limited number of modes.

    When a sparse matrix is passed, and no number of modes is given, all but one
    mode will be calculated. The reason is that the sparse SVD implementation in Scipy
    doesn't allow calculation of all modes. If all modes are required, the user
    must pass a densified version of the matrix (ie. `matrix.toarray()`).

    Parameters
    ----------
    matrix : ndarray or any sparse matrix
        The matrix on which to perform the SVD.
    num_modes : int or None
        The number of singular values and modes to calculate. If this is None,
        and `matrix` is not sparse, all modes will be computed. If this is None and
        `matrix` is sparse, all but one mode will be computed.
    '''
    def __init__(self, matrix=None, num_modes=None, M=None):
        if matrix is None:
            import warnings
            warnings.warn('Deprecated: use "matrix" instead of "M".', DeprecationWarning, stacklevel=2)
            matrix = M

        if matrix is None:
            raise ValueError('You need to supply a matrix.')

        self._matrix = matrix
        self._num_modes = num_modes

        is_sparse = scipy.sparse.issparse(matrix)

        if is_sparse and self.num_modes is None:
            self._num_modes = min(matrix.shape) - 1

        if self.num_modes is None:
            self._svd = np.linalg.svd(matrix, full_matrices=False)
        else:
            self._svd = scipy.sparse.linalg.svds(matrix, int(self.num_modes))

    @property
    def left_singular_modes(self):
        '''The left singular modes of the matrix, as a ModeBasis.
        '''

        return ModeBasis([m for m in self.U.conj().T])

    @property
    def right_singular_modes(self):
        '''The right singular modes of the matrix, as a ModeBasis.
        '''
        from ..mode_basis import ModeBasis

        return ModeBasis([m for m in self.Vt.conj()])

    @property
    def singular_values(self):
        '''The singular values of the matrix.
        '''
        return self.S

    @property
    def U(self):  # noqa: N802
        '''The U matrix of the SVD.
        '''
        return self.svd[0]

    @property
    def S(self):  # noqa: N802
        '''The singular values of the matrix.
        '''
        return self.svd[1]

    @property
    def Vt(self):  # noqa: N802
        '''The V^T matrix of the SVD.
        '''
        return self.svd[2]

    def __getitem__(self, i):
        '''The raw U, S, and V^T matrices of the SVD by index.
        '''
        return self.svd[i]

    @property
    def svd(self):
        '''The raw U, S, and V^T matrices of the SVD as a tuple.
        '''
        return self._svd

    @property
    def num_modes(self):
        '''The number of singular modes that were calculated in this SVD.
        '''
        return self._num_modes

    @property
    def matrix(self):
        '''The matrix for which the SVD was calculated.
        '''
        return self._matrix

    @property
    def M(self):  # noqa: N802
        import warnings
        warnings.warn('Deprecated: use "matrix" instead of "M".', DeprecationWarning, stacklevel=2)
        return self.matrix



import numpy as np

from .fourier_transform import FourierTransform, multiplex_for_tensor_fields, _get_float_and_complex_dtype
from ..field import Field
from ..config import Configuration

class NaiveFourierTransform(FourierTransform):
    '''The naive Fourier transform (NFT).

    This Fourier transform can operate on any types of coordinates, and calculates
    the Fourier integral using brute force. This class is written as generally as
    possible to act as a "ground truth" for the other Fourier transforms. As such,
    it has terrible performance.

    .. note::
        In almost all cases, you should not use this class. The FastFourierTransform and
        MatrixFourierTransform classes, if applicable, are multiple orders of magnitudes faster
        and use multiple orders of magnitude less memory. This class should only be used in a last
        resort and if you know what you are doing.

    .. note::
        The transformation matrices can be very large, quickly overwhelming any computer even
        for relatively small input or output grids.

    Parameters
    ----------
    input_grid : Grid
        The grid that is expected for the input field.
    output_grid : Grid
        The grid that is produced by the Fourier transform.
    precompute_matrices : boolean or None
        Whether to precompute the matrices used in the NFT. Turning this on will cache calculated matrices
        for future evaluations of the Fourier transform. This will use a large amount of memory. If this is
        None, the choice will be determined by the configuration file.

    Raises
    ------
    ValueError
        If the output grid has a different dimension than the input grid.
    '''
    def __init__(self, input_grid, output_grid, precompute_matrices=None):
        if input_grid.ndim != output_grid.ndim:
            raise ValueError('The input_grid must have the same dimensions as the output_grid.')

        self.input_grid = input_grid
        self.output_grid = output_grid

        self._matrix_forward = None
        self._matrix_backward = None

        if precompute_matrices is None:
            precompute_matrices = Configuration().fourier.nft.precompute_matrices
        self.precompute_matrices = precompute_matrices

        self.coords_in = np.array(self.input_grid.as_('cartesian').coords)
        self.coords_out = np.array(self.output_grid.as_('cartesian').coords)

    @property
    def matrix_forward(self):
        '''The cached forward propagation matrix.
        '''
        if not self.precompute_matrices:
            return self.get_transformation_matrix_forward()

        if self._matrix_forward is None:
            self._matrix_forward = self.get_transformation_matrix_forward()

        return self._matrix_forward

    @property
    def matrix_backward(self):
        '''The cached backward propagation matrix.
        '''
        if not self.precompute_matrices:
            return self.get_transformation_matrix_backward()

        if self._matrix_backward is None:
            self._matrix_backward = self.get_transformation_matrix_backward()

        return self._matrix_backward

    @multiplex_for_tensor_fields
    def forward(self, field):
        '''Returns the forward Fourier transform of the :class:`Field` field.

        Parameters
        ----------
        field : Field
            The field to Fourier transform.

        Returns
        --------
        Field
            The Fourier transform of the field.
        '''
        if self.precompute_matrices:
            res = self.matrix_forward.dot(field.ravel())
        else:
            res = np.array([(field * self.input_grid.weights).dot(np.exp(-1j * np.dot(p, self.coords_in))) for p in self.coords_out.T])

        float_dtype, complex_dtype = _get_float_and_complex_dtype(field.dtype)
        return Field(res, self.output_grid).astype(complex_dtype, copy=False)

    @multiplex_for_tensor_fields
    def backward(self, field):
        '''Returns the inverse Fourier transform of the :class:`Field` field.

        Parameters
        ----------
        field : Field
            The field to inverse Fourier transform.

        Returns
        --------
        Field
            The inverse Fourier transform of the field.
        '''
        if self.precompute_matrices:
            res = self.matrix_backward.dot(field.ravel())
        else:
            res = np.array([(field * self.output_grid.weights).dot(np.exp(1j * np.dot(p, self.coords_out))) for p in self.coords_in.T])
            res /= (2 * np.pi)**self.input_grid.ndim

        float_dtype, complex_dtype = _get_float_and_complex_dtype(field.dtype)
        return Field(res, self.input_grid).astype(complex_dtype, copy=False)



import os
import shutil
import time
import sys

from nbclient import NotebookClient
from nbclient.exceptions import CellExecutionError
from nbconvert.exporters import RSTExporter
from nbconvert.writers import FilesWriter
import nbformat

from PIL import Image, ImageChops

def compile_tutorial(tutorial_name, force_recompile=False):
    print('- Tutorial "' + tutorial_name + '"')

    notebook_path = 'tutorial_notebooks/' + tutorial_name + '/' + tutorial_name + '.ipynb'
    export_path = 'tutorials/' + tutorial_name + '/' + tutorial_name
    thumb_dest = os.path.dirname(export_path) + '/thumb.png'

    if not os.path.exists(os.path.dirname(export_path)):
        os.makedirs(os.path.dirname(export_path))

    # Read in notebook
    print('  Reading notebook...')
    notebook = nbformat.read(notebook_path, 4)

    # Scrape title, description and thumbnail
    first_cell = notebook.cells[0]

    title = first_cell.source.splitlines()[0]
    if '#' in title:
        title = title.replace('#', '').strip()

    description = ''
    for line in first_cell.source.splitlines()[1:]:
        if line.strip():
            description = line.strip()
            break

    if not description:
        print('  Description could not be found in the notebook.')

    if 'thumbnail_figure_index' in notebook.metadata:
        thumbnail_figure_index = notebook.metadata['thumbnail_figure_index']
    else:
        thumbnail_figure_index = -1

    if 'level' in notebook.metadata:
        level = notebook.metadata['level'].capitalize()
    elif 'difficulty' in notebook.metadata:
        level = notebook.metadata['difficulty'].capitalize()
    else:
        level = 'Unknown'

    # Check if the tutorial was already executed.
    if os.path.exists(export_path + '.rst'):
        if os.path.getmtime(export_path + '.rst') > os.path.getmtime(notebook_path):
            if force_recompile:
                print('  Already compiled. Recompiling anyway...')
            else:
                print('  Already compiled. Skipping compilation...')
                return title, level, description, thumb_dest.split('/', 1)[-1], False

    # Execute notebook if not already executed
    already_executed = any(c.get('outputs') or c.get('execution_count') for c in notebook.cells if c.cell_type == 'code')

    resources = {'metadata': {'path': os.path.dirname(notebook_path)}}

    with_errors = False

    if not already_executed:
        print('  Executing', end='')
        start = time.time()

        additional_cell_1 = {
            "cell_type": "code",
            "execution_count": None,
            "metadata": {},
            "outputs": [],
            "source": r"%matplotlib inline" + '\n' +
                r"%config InlineBackend.print_figure_kwargs = {'bbox_inches': None}"
            }

        additional_cell_2 = {
            "cell_type": "code",
            "execution_count": None,
            "metadata": {},
            "outputs": [],
            "source": "import matplotlib as mpl\nmpl.rcParams['figure.figsize'] = (8, 6)\nmpl.rcParams['figure.dpi'] = 150\nmpl.rcParams['savefig.dpi'] = 150"
            }

        notebook.cells.insert(1, nbformat.from_dict(additional_cell_1))
        notebook.cells.insert(2, nbformat.from_dict(additional_cell_2))

        client = NotebookClient(nb=notebook, resources=resources, timeout=585, kernel_name='python3')

        try:
            with client.setup_kernel():
                for i, cell in enumerate(notebook.cells):
                    print('.', end='')

                    client.execute_cell(cell, i)

            client.set_widgets_metadata()
        except CellExecutionError as err:
            print('  Error while processing notebook:')
            print('  ', err)

            with_errors = True

        print('')

        notebook.cells.pop(2)
        notebook.cells.pop(1)

        end = time.time()

        time_taken = end - start
        if time_taken > 60:
            print('  Execution took %dm%02ds.' % (time_taken / 60, time_taken % 60))
        else:
            print('  Execution took %ds.' % time_taken)
    else:
        print('  Notebook was already executed.')

    print('  Rendering tutorial...')
    exporter = RSTExporter()
    output, resources = exporter.from_notebook_node(notebook, resources)

    writer = FilesWriter(build_directory=os.path.dirname(export_path))
    writer.write(output, resources, notebook_name=os.path.basename(export_path))

    pictures = sorted(resources['outputs'], key=output.find)

    try:
        thumbnail_source = pictures[thumbnail_figure_index]

        # Read in thumbnail source image
        img = Image.open(os.path.dirname(export_path) + '/' + thumbnail_source)

        # Trim whitespace
        bg = Image.new(img.mode, img.size, img.getpixel((0, 0)))
        diff = ImageChops.difference(img, bg)
        diff = ImageChops.add(diff, diff)
        bbox = diff.getbbox()
        if bbox:
            img = img.crop(bbox)

        # Resize image to have a width of 400px
        img.thumbnail([400, 1000])

        # Save thumbnail
        img.save(thumb_dest)
    except:
        shutil.copyfile('_static/no_thumb.png', thumb_dest)

    print('  Done!')

    return title, level, description, thumb_dest.split('/', 1)[-1], with_errors

index_preamble = '''
Tutorials
=========

These tutorials demonstrate the features of HCIPy in the context of a standard workflow. Tutorials are separated in three categories, depending on the required level of familiarity with HCIPy.
'''

entry_template = '''
.. only:: html

    .. container:: tutorial_item

        :doc:`{title} <{name}/{name}>`

        .. container:: tutorial_row

            .. container:: tutorial_thumbnail

                .. figure:: {thumbnail_file}

            .. container:: tutorial_description

                **Level:** {level}

                **Description:** {description}
'''

beginner_preamble = '''
Beginner
--------

These tutorials provide an introduction to the basic parts of HCIPy. New users should read these tutorials to get started with HCIPy.
'''

intermediate_preamble = '''
Intermediate
------------

These tutorials show the main functionality using the built-in classes of HCIPy. These tutorials focus on one aspect of high-contrast imaging.
'''

advanced_preamble = '''
Advanced
--------
These tutorials show how to use HCIPy for your own research. This includes extending HCIPy with your own optical elements and advanced use cases.
'''

expert_preamble = '''
Expert
------

These tutorials provide examples from actual published research that made heavy use of HCIPy for their optical propagations.
'''

unknown_preamble = '''
Unknown
-------

These tutorials do not have their level of difficulty rated.
'''

level_preambles = [beginner_preamble, intermediate_preamble, advanced_preamble, expert_preamble, unknown_preamble]

def compile_all_tutorials():
    print('Compiling all tutorials...')

    tutorials = {}
    tutorial_names = sorted(os.listdir('tutorial_notebooks/'))
    tutorial_names = [name for name in tutorial_names if 'checkpoint' not in name]

    with_errors = False

    for name in tutorial_names:
        tutorials[name] = compile_tutorial(name)

        with_errors = with_errors or tutorials[name][-1]

    # Sort by level
    levels = ['Beginner', 'Intermediate', 'Advanced', 'Expert', 'Unknown']

    tutorial_names_by_level = [[] for i in range(len(levels))]
    for name in tutorial_names:
        tutorial_names_by_level[levels.index(tutorials[name][1])].append(name)

    with open('tutorials/index.rst', 'w') as f:
        f.write(index_preamble)

        for preamble, names in zip(level_preambles, tutorial_names_by_level):
            # Don't write this level if there are no tutorials in it.
            if len(names) == 0:
                continue

            f.write(preamble)

            # Write toctree
            f.write('\n.. toctree::\n    :maxdepth: 1\n    :hidden:\n\n')
            for name in names:
                f.write('    ' + name + '/' + name + '\n')
            f.write('\n\n')

            # Write list
            for name in names:
                title, level, desc, thumb, _ = tutorials[name]
                f.write(entry_template.format(thumbnail_file=thumb, title=title, level=level, description=desc, name=name))

    return with_errors

if __name__ == '__main__':
    # Compile all tutorials
    import os
    os.chdir(os.path.dirname(os.path.abspath(__file__)))

    import sys
    sys.path.insert(0, '.')

    with_errors = compile_all_tutorials()

    if with_errors:
        sys.exit(1)



import numpy as np

from .grid import Grid
from .coordinates import UnstructuredCoords

class PolarGrid(Grid):
    '''A grid representing a two-dimensional Polar coordinate system.
    '''
    _coordinate_system = 'polar'

    @property
    def r(self):
        '''The radial coordinate (dimension 0).
        '''
        return self.coords[0]

    @property
    def theta(self):
        ''' The angular coordinate (dimension 1).
        '''
        return self.coords[1]

    def scale(self, scale):
        '''Scale the grid in-place.

        Parameters
        ----------
        scale : array_like
            The factor with which to scale the grid.

        Returns
        -------
        Grid
            Itself to allow for chaining these transformations.
        '''
        self.coords *= np.array([scale, 1])
        self.weights *= np.abs(scale)**(self.ndim)
        return self

    def shift(self, shift):
        '''Shift the grid in-place.

        .. caution::
            All structure in the coordinates will be destroyed.

        Parameters
        ----------
        shift : array_like
            The amount with which to shift the grid.

        Returns
        -------
        Grid
            Itself to allow for chaining these transformations.
        '''
        new_grid = PolarGrid(UnstructuredCoords(self.coords)).as_('cartesian')
        self.coords = new_grid.shifted(shift).as_('polar').coords

        return self

    def shifted(self, shift):
        '''A shifted copy of this grid.

        .. caution::
            The returned grid is a Cartesian grid.

        Parameters
        ----------
        shift : array_like
            The amount with which to shift the grid.

        Returns
        -------
        Grid
            The scaled grid.
        '''
        grid = self.as_('cartesian')
        grid.shift(shift)
        return grid

    def rotate(self, angle, axis=None):
        '''Rotate the grid in-place.

        Parameters
        ----------
        angle : scalar
            The angle in radians.
        axis : ndarray or None
            The axis of rotation. For this (polar) grid, it is ignored.

        Returns
        -------
        Grid
            Itself to allow for chaining these transformations.
        '''
        self.theta += angle
        return self

    @staticmethod
    def _get_automatic_weights(coords):
        return None

def _cartesian_to_polar(self):
    from .coordinates import UnstructuredCoords

    x = self.x
    y = self.y
    r = np.hypot(x, y)
    theta = np.arctan2(y, x)
    return PolarGrid(UnstructuredCoords([r, theta]))

def _polar_to_cartesian(self):
    from .coordinates import UnstructuredCoords
    from .cartesian_grid import CartesianGrid

    r = self.r
    theta = self.theta
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return CartesianGrid(UnstructuredCoords([x, y]))

Grid._add_coordinate_system('polar', PolarGrid)

Grid._add_coordinate_system_transformation('cartesian', 'polar', _cartesian_to_polar)
Grid._add_coordinate_system_transformation('polar', 'cartesian', _polar_to_cartesian)



from scipy.interpolate import RegularGridInterpolator, LinearNDInterpolator
import numpy as np
from ..field import Field

def make_linear_interpolator_separated(field, grid=None, fill_value=np.nan):
    '''Make a linear interpolator for a Field on a separated grid.

    Parameters
    ----------
    field : Field or ndarray
        The field to interpolate.
    grid : Grid or None
        The grid of the field. If it is given, the grid of `field` is replaced by this grid.
    fill_value : scalar
        The value to use for points outside of the domain of the input field. If this is None, the values
        outside the domain are extrapolated.

    Returns
    -------
    Field generator
        The interpolator, as a Field generator. The grid on which this field generator will evaluated, does
        not have to be separated.
    '''
    if grid is None:
        grid = field.grid
    else:
        field = Field(field, grid)

    axes_reversed = np.array(grid.separated_coords)
    interp = RegularGridInterpolator(axes_reversed, np.asarray(field.shaped), 'linear', False, fill_value)

    def interpolator(evaluated_grid):
        evaluated_coords = np.flip(np.array(evaluated_grid.coords), 0)
        res = interp(evaluated_coords.T)
        return Field(res.ravel(), evaluated_grid)

    return interpolator

def make_linear_interpolator_unstructured(field, grid=None, fill_value=np.nan):
    '''Make a linear interpolator for a Field on an unstructured grid.

    Parameters
    ----------
    field : Field or array_like
        The field to interpolate.
    grid : Grid or None
        The grid of the field. If it is given, the grid of `field` is replaced by this grid.
    fill_value : scalar
        The value to use for points outside of the domain of the input field. Extrapolation is not supported.

    Returns
    -------
    Field generator
        The interpolator as a Field generator. The grid on which this field generator will be evaluated does
        not need to have any structure.
    '''
    if fill_value is None:
        raise ValueError('Extrapolation is not supported for a linear interpolator on an unstructured grid.')

    if grid is None:
        grid = field.grid
    else:
        field = Field(field, grid)

    interp = LinearNDInterpolator(grid.points, np.asarray(field), fill_value)

    def interpolator(evaluated_grid):
        res = interp(evaluated_grid.points)
        return Field(res, evaluated_grid)

    return interpolator

def make_linear_interpolator(field, grid=None, fill_value=None):
    '''Make a linear interpolator for a Field on any type of grid.

    Parameters
    ----------
    field : Field or array_like
        The field to interpolate.
    grid : Grid or None
        The grid of the field. If it is given, the grid of `field` is replaced by this grid.
    fill_value : scalar or None
        The value to use for points outside of the domain of the input field. Extrapolation is not supported.
        If it is None, a numpy.nan value will be used for points outside of the domain.

    Returns
    -------
    Field generator
        The interpolator as a Field generator. The grid on which this field generator will be evaluated does
        not need to have any structure.
    '''
    if grid is None:
        grid = field.grid

    if grid.is_unstructured:
        return make_linear_interpolator_unstructured(field, grid, fill_value)
    else:
        return make_linear_interpolator_separated(field, grid, fill_value)
